﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Windows.Media.Imaging;

namespace PerlinNoiseShader.ImprovedNoise2D
{
    /// <summary>
    /// consists of methods to calculate single ocatves of improved 2D Perlin noise
    /// </summary>
    public class Noise2D:NoiseMask
    {
        private ImageBrush oGrad1Brush;
        private ImageBrush oGrad2Brush;
        private ImageBrush oGradBrush;
        private ImageBrush oGBrush;
        private Noise2DEffect[] aNoise2DEffect = new Noise2DEffect[8];
        public Noise2D():this((new Random()).Next()){}
        public Noise2D(int random){
            oDrawRectangle = new Rectangle();
            iUIElement = oDrawRectangle;
            for (int i = 0; i < 8; i++)
            {
                aNoise2DEffect[i] = new Noise2DEffect((Noise2DEffect.KindOf)i);
            }
            oGBrush = new ImageBrush();
            getG();
            oGradBrush = new ImageBrush();
            oGrad1Brush = new ImageBrush();
            oGrad2Brush = new ImageBrush();
            reInitialize(random);
        }

        /// <summary>
        /// get the gradients
        /// </summary>
        private void getG()
        {
            WriteableBitmap oGradBitmap = new WriteableBitmap(8, 1);
            for (int i = 0; i < 8; i++)
            {
                int h = i & 7;
                int x, y;
                if (h < 4)
                {
                    x = ((h & 1) == 1) ? -1 : 1;
                    y = ((h & 2) == 2) ? -2 : 2;
                }
                else
                {
                    y = ((h & 1) == 1) ? -1 : 1;
                    x = ((h & 2) == 2) ? -2 : 2;
                }
                x += 2;
                y += 2;
                oGradBitmap.Pixels[i] = (255 << 24) | (x << 16) | (y << 8);
            }
            oGradBitmap.Invalidate();
            oGBrush.ImageSource = oGradBitmap;
        }

        /// <summary>
        /// initializes the perlin noise for a new random seed value
        /// </summary>
        public void reInitialize()
        {
            reInitialize((new Random()).Next());
        }

        /// <summary>
        /// initialized the perlin noise for a new random seed value
        /// </summary>
        /// <param name="random">random seed</param>
        public void reInitialize(int random)
        {
            oRandom = new Random(random);
            thisSeed = random;
            oPBrush = getP();
            Rectangle oRectangle = new Rectangle();
            oRectangle.Width = 256;
            oRectangle.Height = 256;
            oRectangle.Fill = new SolidColorBrush(Color.FromArgb(255, 255, 0, 0));
            PrepNoise2DEffect oPrepNoise2DEffect = new PrepNoise2DEffect(PrepNoise2DEffect.Prep.GRAD1);
            oPrepNoise2DEffect.perm = oPBrush;
            oPrepNoise2DEffect.grad = oGBrush;
            oRectangle.Effect = oPrepNoise2DEffect;
            WriteableBitmap grad1Bitmap = new WriteableBitmap(256, 256);
            grad1Bitmap.Render(oRectangle, null);
            grad1Bitmap.Invalidate();
            oPrepNoise2DEffect = new PrepNoise2DEffect(PrepNoise2DEffect.Prep.GRAD2);
            oPrepNoise2DEffect.perm = oPBrush;
            oPrepNoise2DEffect.grad = oGBrush;
            oRectangle.Effect = oPrepNoise2DEffect;
            WriteableBitmap grad2Bitmap = new WriteableBitmap(256, 256);
            grad2Bitmap.Render(oRectangle, null);
            grad2Bitmap.Invalidate();
            oPrepNoise2DEffect = new PrepNoise2DEffect(PrepNoise2DEffect.Prep.GRAD);
            oPrepNoise2DEffect.perm = oPBrush;
            oPrepNoise2DEffect.grad = oGBrush;
            oRectangle.Effect = oPrepNoise2DEffect;
            WriteableBitmap gradBitmap = new WriteableBitmap(256, 256);
            gradBitmap.Render(oRectangle, null);
            gradBitmap.Invalidate();
            
            oGradBrush.ImageSource = gradBitmap;
            oGrad1Brush.ImageSource = grad1Bitmap;
            oGrad2Brush.ImageSource = grad2Bitmap;
        }

        /// <summary>
        /// calculates a single octave of 2D improved perlin noise
        /// </summary>
        /// <param name="freq">frequency at the x and y axis</param>
        /// <param name="offset">left upper corner of the plane</param>
        /// <param name="amplitude"></param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void singleNoise(Size freq, Size offset, float amplitude,bool absolute, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            Size testSize = new Size(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight);
            if (!thisSize.Equals(testSize))
            {
                thisSize = testSize;
                if (oDrawRectangle != null) createRectangle();
            }

            Noise2DEffect oNoise2DEffect = absolute ? aNoise2DEffect[1] : aNoise2DEffect[0];
            iUIElement.Effect = oNoise2DEffect;
            oNoise2DEffect.amplitude = absolute ? amplitude : 0.5f * amplitude;
            oNoise2DEffect.c1 = c1;
            oNoise2DEffect.c2 = c2;
            oNoise2DEffect.size = new Size(freq.Width * thisSize.Width, freq.Height * thisSize.Height);
            oNoise2DEffect.offset = new Size(offset.Width * freq.Width, offset.Height * freq.Height);
            oNoise2DEffect.grad1 = oGrad1Brush;
            oNoise2DEffect.grad2 = oGrad2Brush;
            oWriteableBitmap.Render(iUIElement, oCustomTransform);
            oWriteableBitmap.Invalidate();
        }

        /// <summary>
        /// calculates a single octave of 2D improved tileable perlin noise
        /// </summary>
        /// <param name="freq">frequency at the x and y axis</param>
        /// <param name="tile">if freq=(i/width, j/height), the noise tiles at the borders for i=2..256 and j=2..256</param>
        /// <param name="offset">left upper corner of the plane</param>
        /// <param name="amplitude"></param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void singleNoise(Size freq, iSize tile, Size offset, float amplitude, bool absolute, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            Size testSize = new Size(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight);
            if (!thisSize.Equals(testSize))
            {
                thisSize = testSize;
                if (oDrawRectangle != null) createRectangle();
            }

            Noise2DEffect oNoise2DEffect = absolute ? aNoise2DEffect[3] : aNoise2DEffect[2];
            iUIElement.Effect = oNoise2DEffect;
            oNoise2DEffect.amplitude = absolute ? amplitude : 0.5f * amplitude;
            oNoise2DEffect.c1 = c1;
            oNoise2DEffect.c2 = c2;
            oNoise2DEffect.size = new Size(freq.Width * thisSize.Width, freq.Height * thisSize.Height);
            oNoise2DEffect.tile = tile;
            oNoise2DEffect.offset = new Size(offset.Width * freq.Width, offset.Height * freq.Height);
            oNoise2DEffect.grad = oGradBrush;
            oWriteableBitmap.Render(iUIElement, oCustomTransform);
            oWriteableBitmap.Invalidate();
        }

        /// <summary>
        /// calculates a single octave of 2D improved perlin noise
        /// </summary>
        /// <param name="freq">frequency at the x and y axis</param>
        /// <param name="offset">left upper corner of the plane</param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void noise(Size freq, Size offset, bool absolute, WriteableBitmap oWriteableBitmap)
        {
            Size testSize = new Size(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight);
            if (!thisSize.Equals(testSize))
            {
                thisSize = testSize;
                if (oDrawRectangle != null) createRectangle();
            }

            Noise2DEffect oNoise2DEffect = absolute ? aNoise2DEffect[5] : aNoise2DEffect[4];
            iUIElement.Effect = oNoise2DEffect;
            oNoise2DEffect.size = new Size(freq.Width * thisSize.Width, freq.Height * thisSize.Height);
            oNoise2DEffect.offset = new Size(offset.Width * freq.Width, offset.Height * freq.Height);
            oNoise2DEffect.grad1 = oGrad1Brush;
            oNoise2DEffect.grad2 = oGrad2Brush;
            oWriteableBitmap.Render(iUIElement, oCustomTransform);
            oWriteableBitmap.Invalidate();
        }

        /// <summary>
        /// calculates a single octave of 2D improved tileable perlin noise to be mixed with other noise
        /// </summary>
        /// <param name="freq">frequency at the x and y axis</param>
        /// <param name="tile">if freq=(i/width, j/height), the noise tiles at the borders for i=2..256 and j=2..256</param>
        /// <param name="offset">left upper corner of the plane</param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void noise(Size freq, iSize tile, Size offset, bool absolute, WriteableBitmap oWriteableBitmap)
        {
            Size testSize = new Size(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight);
            if (!thisSize.Equals(testSize))
            {
                thisSize = testSize;
                if (oDrawRectangle != null) createRectangle();
            }

            Noise2DEffect oNoise2DEffect = absolute ? aNoise2DEffect[7] : aNoise2DEffect[6];
            iUIElement.Effect = oNoise2DEffect;
            oNoise2DEffect.size = new Size(freq.Width * thisSize.Width, freq.Height * thisSize.Height);
            oNoise2DEffect.tile = tile;
            oNoise2DEffect.offset = new Size(offset.Width * freq.Width, offset.Height * freq.Height);
            oNoise2DEffect.grad = oGradBrush;
            oWriteableBitmap.Render(iUIElement, oCustomTransform);
            oWriteableBitmap.Invalidate();
        }

    }
}
