﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Windows.Media.Media3D;

namespace PerlinNoiseShader
{
    public class Matrix3d
    {
        public double m00, m01, m02;
        public double m10, m11, m12;
        public double m20, m21, m22;

        /**
         * Constructor instantiates a new <code>Matrix3f</code> object. The
         * initial values for the matrix is that of the identity matrix.
         *  
         */
        public Matrix3d()
        {
            loadIdentity();
        }

        /**
         * constructs a matrix with the given values.
         * 
         * @param m00
         *            0x0 in the matrix.
         * @param m01
         *            0x1 in the matrix.
         * @param m02
         *            0x2 in the matrix.
         * @param m10
         *            1x0 in the matrix.
         * @param m11
         *            1x1 in the matrix.
         * @param m12
         *            1x2 in the matrix.
         * @param m20
         *            2x0 in the matrix.
         * @param m21
         *            2x1 in the matrix.
         * @param m22
         *            2x2 in the matrix.
         */
        public Matrix3d(float m00, float m01, float m02, float m10, float m11,
                float m12, float m20, float m21, float m22)
        {

            this.m00 = m00;
            this.m01 = m01;
            this.m02 = m02;
            this.m10 = m10;
            this.m11 = m11;
            this.m12 = m12;
            this.m20 = m20;
            this.m21 = m21;
            this.m22 = m22;
        }

        /**
         * Copy constructor that creates a new <code>Matrix3f</code> object that
         * is the same as the provided matrix.
         * 
         * @param mat
         *            the matrix to copy.
         */
        public Matrix3d(Matrix3d mat)
        {
            copy(mat);
        }

        /**
         * <code>copy</code> transfers the contents of a given matrix to this
         * matrix. If a null matrix is supplied, this matrix is set to the identity
         * matrix.
         * 
         * @param matrix
         *            the matrix to copy.
         */
        public void copy(Matrix3d matrix)
        {
            if (null == matrix)
            {
                loadIdentity();
            }
            else
            {
                m00 = matrix.m00;
                m01 = matrix.m01;
                m02 = matrix.m02;
                m10 = matrix.m10;
                m11 = matrix.m11;
                m12 = matrix.m12;
                m20 = matrix.m20;
                m21 = matrix.m21;
                m22 = matrix.m22;
            }
        }

        
        public void loadIdentity()
        {
            m01 = m02 = m10 = m12 = m20 = m21 = 0;
            m00 = m11 = m22 = 1;
        }

        public void fromAngleAxis(double angle, Vector3d axis)
        {
            if (axis.x == 0 && axis.y == 0 && axis.z == 0)
            {
                loadIdentity();
                return;
            }

            Vector3d normAxis = axis.normalize();
            fromAngleNormalAxis(angle, normAxis);
        }
        public void fromAngleNormalAxis(double angle, Vector3d axis)
        {
            double fCos = Math.Cos(angle);
            double fSin = Math.Sin(angle);
            double fOneMinusCos = ((double)1.0) - fCos;
            double fX2 = axis.x * axis.x;
            double fY2 = axis.y * axis.y;
            double fZ2 = axis.z * axis.z;
            double fXYM = axis.x * axis.y * fOneMinusCos;
            double fXZM = axis.x * axis.z * fOneMinusCos;
            double fYZM = axis.y * axis.z * fOneMinusCos;
            double fXSin = axis.x * fSin;
            double fYSin = axis.y * fSin;
            double fZSin = axis.z * fSin;

            m00 = fX2 * fOneMinusCos + fCos;
            m01 = fXYM - fZSin;
            m02 = fXZM + fYSin;
            m10 = fXYM + fZSin;
            m11 = fY2 * fOneMinusCos + fCos;
            m12 = fYZM - fXSin;
            m20 = fXZM - fYSin;
            m21 = fYZM + fXSin;
            m22 = fZ2 * fOneMinusCos + fCos;
        }

        public Vector3d mult(Vector3d vec)
        {
            Vector3d product = new Vector3d();
            
            double x = vec.x;
            double y = vec.y;
            double z = vec.z;

            product.x = m00 * x + m01 * y + m02 * z;
            product.y = m10 * x + m11 * y + m12 * z;
            product.z = m20 * x + m21 * y + m22 * z;
            return product;
        }

        public Matrix3d mult(Matrix3d mat)
        {

            double temp00, temp01, temp02;
            double temp10, temp11, temp12;
            double temp20, temp21, temp22;

            Matrix3d product = new Matrix3d();
            temp00 = m00 * mat.m00 + m01 * mat.m10 + m02 * mat.m20;
            temp01 = m00 * mat.m01 + m01 * mat.m11 + m02 * mat.m21;
            temp02 = m00 * mat.m02 + m01 * mat.m12 + m02 * mat.m22;
            temp10 = m10 * mat.m00 + m11 * mat.m10 + m12 * mat.m20;
            temp11 = m10 * mat.m01 + m11 * mat.m11 + m12 * mat.m21;
            temp12 = m10 * mat.m02 + m11 * mat.m12 + m12 * mat.m22;
            temp20 = m20 * mat.m00 + m21 * mat.m10 + m22 * mat.m20;
            temp21 = m20 * mat.m01 + m21 * mat.m11 + m22 * mat.m21;
            temp22 = m20 * mat.m02 + m21 * mat.m12 + m22 * mat.m22;

            product.m00 = temp00;
            product.m01 = temp01;
            product.m02 = temp02;
            product.m10 = temp10;
            product.m11 = temp11;
            product.m12 = temp12;
            product.m20 = temp20;
            product.m21 = temp21;
            product.m22 = temp22;

            return product;
        }

        public Matrix3D getMatrix3D()
        {
            Matrix3D oMatrix3D = new Matrix3D();
            oMatrix3D.M11 = this.m00;
            oMatrix3D.M12 = this.m01;
            oMatrix3D.M13 = this.m02;
            oMatrix3D.M21 = this.m10;
            oMatrix3D.M22 = this.m11;
            oMatrix3D.M23 = this.m12;
            oMatrix3D.M31 = this.m20;
            oMatrix3D.M32 = this.m21;
            oMatrix3D.M33 = this.m22;
            return oMatrix3D;
        }
    }
}
