﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Windows.Media.Imaging;
using PerlinNoiseShader.PerlinNoise2D;
using PerlinNoiseShader.Simplex2D;
using PerlinNoiseShader.ImprovedNoise;
using PerlinNoiseShader.ImprovedNoise2D;

namespace PerlinNoiseShader.NoiseBlender
{
    /// <summary>
    /// generates perlin noise up to 16 octaves. the array size of the noise methods parameters must match the octave count. 
    /// At "turbulence" methods, the array parameters can be null. 
    /// In the perlin noise classes, a single octave is created. this is called many times with different frequencies and is mixed together. 
    /// there is sum(noise) and sum(|noise|). The sum(|noise|) can be a parameter of a cosine function.
    /// </summary>
    public class NoiseBlenderClass
    {

        #region variables

        private int octaves = -1;
        private int octavesCos = -1;
        private int seed = -1;
        private WriteableBitmap[] aWriteableBitmap;
        private ImageBrush[] aImageBrush;
        private PerlinNoise2DClass oPerlinNoise2DClass = null;
        private Noise2D oNoise2D = null;
        private Simplex2DClass oSimplex2DClass = null;
        private ImprovedNoiseClass oImprovedNoiseClass = null;
        private NoiseBlenderEffect oNoiseBlenderEffect = null;
        private NoiseBlenderCosineEffect oNoiseBlenderCosineEffect = null;
        private NoiseBlenderCosineEffect3D oNoiseBlenderCosineEffect3D = null;
        private UIElement iUIElement = null;
        private Transform oTransform = null;

        #endregion

        #region properties
        public int currentSeed
        {
            get { return seed; }
        }
        
        /// <summary>
        /// custom UIElement for custom shape
        /// </summary>
        public UIElement customUIElement
        {
            get { return iUIElement; }
            set
            {
                iUIElement = value;
                if (oPerlinNoise2DClass != null)
                {
                    oPerlinNoise2DClass.customUIElement = iUIElement;
                }
                if (oSimplex2DClass != null)
                {
                    oSimplex2DClass.customUIElement = iUIElement;
                }
                if (oImprovedNoiseClass != null)
                {
                    oImprovedNoiseClass.customUIElement = iUIElement;
                }
                if (oNoise2D != null)
                {
                    oNoise2D.customUIElement = iUIElement;
                }
            }
        }

        /// <summary>
        /// custom 2D transformation
        /// </summary>
        public Transform oCustomTransform
        {
            get { return oTransform; }
            set
            {
                oTransform = value;
                if (oPerlinNoise2DClass != null)
                {
                    oPerlinNoise2DClass.oCustomTransform = oTransform;
                }
                if (oNoise2D != null)
                {
                    oNoise2D.oCustomTransform = oTransform;
                }
                if (oSimplex2DClass != null)
                {
                    oSimplex2DClass.oCustomTransform = oTransform;
                }
                if (oImprovedNoiseClass != null)
                {
                    oImprovedNoiseClass.oCustomTransform = oTransform;
                }
            }
        }

        #endregion

        #region standard 2D Perlin noise
        /// <summary>
        /// calculates sum(noise) or sum(|noise|) with a random seed value
        /// </summary>
        /// <param name="freqXY">frequencies at the x and y axis</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="tileable">true, if it tiles at the WriteableBitmap borders</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destintation bitmap</param>
        public void noises(Size[] freqXY, int octaves, Size[] aOffset, float[] aAmplitude, bool absolute, bool tileable, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            noises(freqXY, octaves, (new Random()).Next(), aOffset, aAmplitude, absolute, tileable, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates sum(noise) or sum(|noise|) with a random seed value
        /// </summary>
        /// <param name="freqXY">frequencies at the x and y axis</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="tileableX">true, if tileable at the width</param>
        /// <param name="tileableY">true, if tileable at the height</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destintation bitmap</param>
        public void noises(Size[] freqXY, int octaves, Size[] aOffset, float[] aAmplitude, bool absolute, bool tileableX, bool tileableY, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            noises(freqXY, octaves, (new Random()).Next(), aOffset, aAmplitude, absolute, tileableX, tileableY, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates amplitude*cos(x*xyMult.x+y*xyMult.y+sum(|noise|)+shift) with a random seed value
        /// </summary>
        /// <param name="freqXY">frequencies at the x and y axis</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="amplitude">amplitude of the cosine function</param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="tileable">true, if tileable</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void noises(Size[] freqXY, int octaves, Size[] aOffset, float[] aAmplitude, float amplitude, Point xyMult, float shift, bool tileable, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            noises(freqXY, octaves, (new Random()).Next(), aOffset, aAmplitude, amplitude, xyMult, shift, tileable, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates amplitude*cos(x*xyMult.x+y*xyMult.y+sum(|noise|)+shift) with a random seed value
        /// </summary>
        /// <param name="freqXY">frequencies at the x and y axis</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="amplitude">amplitude of the cosine function</param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="tileableX">true, if tiles at the width</param>
        /// <param name="tileableY">true, if tiles at the height</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void noises(Size[] freqXY, int octaves, Size[] aOffset, float[] aAmplitude, float amplitude, Point xyMult, float shift, bool tileableX, bool tileableY, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            noises(freqXY, octaves, (new Random()).Next(), aOffset, aAmplitude, amplitude, xyMult, shift, tileableX, tileableY, c1, c2, oWriteableBitmap);
        }
        
        /// <summary>
        /// calculates sum(noise) or sum(|noise|)
        /// </summary>
        /// <param name="freqXY">frequencies at the x and y axis</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="tileable">true, if it tiles at the WriteableBitmap borders</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destintation bitmap</param>
        public void noises(Size[] freqXY, int octaves, int seed, Size[] aOffset, float[] aAmplitude, bool absolute,bool tileable, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if(aOffset.Length!=octaves) return;
            if(aAmplitude.Length!=octaves) return;
            if (freqXY.Length != octaves) return;
            bool isNotEqualSize = oPerlinNoise2DClass == null ? true : !(new Size(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight)).Equals(oPerlinNoise2DClass.thisSize);
            if (octaves != this.octaves)
            {
                this.octaves = octaves;
                if (octaves > 1)
                {
                    oNoiseBlenderEffect = new NoiseBlenderEffect(octaves - 1, absolute);
                    aWriteableBitmap = new WriteableBitmap[octaves];
                    aImageBrush = new ImageBrush[octaves];
                    for (int i = 0; i < octaves; i++)
                    {
                        aWriteableBitmap[i] = new WriteableBitmap(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight);
                        aImageBrush[i] = new ImageBrush();
                        aImageBrush[i].ImageSource = aWriteableBitmap[i];
                    }
                    oNoiseBlenderEffect.aInput = aImageBrush;
                }
                else oNoiseBlenderEffect = null;

            }
            
                
            if (oPerlinNoise2DClass == null)
            {
                this.seed = seed;
                oPerlinNoise2DClass = new PerlinNoise2DClass(seed);
                oPerlinNoise2DClass.customUIElement = iUIElement;
                oPerlinNoise2DClass.oCustomTransform = oTransform;
            }else if (this.seed!=seed){
                this.seed = seed;
                oPerlinNoise2DClass.reInitialize(seed);
            }
            if (octaves == 1)
            {
                if (tileable)
                {
                    oPerlinNoise2DClass.singleTileableNoise(freqXY[0], aOffset[0], aAmplitude[0], absolute, c1, c2, oWriteableBitmap);
                }
                else
                {
                    oPerlinNoise2DClass.singleNoise(freqXY[0], aOffset[0], aAmplitude[0], absolute, c1, c2, oWriteableBitmap);
                }
            }
            else
            {
                if (oNoiseBlenderEffect.absolute != absolute)
                {
                    oNoiseBlenderEffect = new NoiseBlenderEffect(octaves - 1, absolute);
                    oNoiseBlenderEffect.aInput = aImageBrush;
                }
                oNoiseBlenderEffect.c1 = c1;
                oNoiseBlenderEffect.c2 = c2;
                oNoiseBlenderEffect.aAmplitude = aAmplitude;
                for (int i = 0; i < octaves; i++)
                {
                    if (tileable)
                    {
                        oPerlinNoise2DClass.tileableNoise(freqXY[i], aOffset[i], absolute, aWriteableBitmap[i]);
                    }
                    else
                    {
                        oPerlinNoise2DClass.noise(freqXY[i], aOffset[i], absolute, aWriteableBitmap[i]);
                    }
                }
                oPerlinNoise2DClass.currentUIElement.Effect = oNoiseBlenderEffect;
                oWriteableBitmap.Render(oPerlinNoise2DClass.currentUIElement, oTransform);
                oWriteableBitmap.Invalidate();
                
            }
        }

        /// <summary>
        /// calculates sum(noise) or sum(|noise|)
        /// </summary>
        /// <param name="freqXY">frequencies at the x and y axis</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="tileableX">true, if tileable at the width</param>
        /// <param name="tileableY">true, if tileable at the height</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destintation bitmap</param>
        public void noises(Size[] freqXY, int octaves, int seed, Size[] aOffset, float[] aAmplitude, bool absolute, bool tileableX, bool tileableY, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset.Length != octaves) return;
            if (aAmplitude.Length != octaves) return;
            if (freqXY.Length != octaves) return;
            bool isNotEqualSize = oPerlinNoise2DClass == null ? true : !(new Size(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight)).Equals(oPerlinNoise2DClass.thisSize);
            if (octaves != this.octaves || isNotEqualSize)
            {
                this.octaves = octaves;
                if (octaves > 1)
                {
                    oNoiseBlenderEffect = new NoiseBlenderEffect(octaves - 1, absolute);
                    aWriteableBitmap = new WriteableBitmap[octaves];
                    aImageBrush = new ImageBrush[octaves];
                    for (int i = 0; i < octaves; i++)
                    {
                        aWriteableBitmap[i] = new WriteableBitmap(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight);
                        aImageBrush[i] = new ImageBrush();
                        aImageBrush[i].ImageSource = aWriteableBitmap[i];
                    }
                    oNoiseBlenderEffect.aInput = aImageBrush;
                }
                else oNoiseBlenderEffect = null;

            }


            if (oPerlinNoise2DClass == null)
            {
                this.seed = seed;
                oPerlinNoise2DClass = new PerlinNoise2DClass(seed);
                oPerlinNoise2DClass.customUIElement = iUIElement;
                oPerlinNoise2DClass.oCustomTransform = oTransform;
            }
            else if (this.seed != seed)
            {
                this.seed = seed;
                oPerlinNoise2DClass.reInitialize(seed);
            }
            if (octaves == 1)
            {
                if (tileableX ||tileableY)
                {
                    oPerlinNoise2DClass.singleTileableNoise(freqXY[0], aOffset[0], aAmplitude[0], absolute,tileableX,tileableY, c1, c2, oWriteableBitmap);
                }
                else
                {
                    oPerlinNoise2DClass.singleNoise(freqXY[0], aOffset[0], aAmplitude[0], absolute, c1, c2, oWriteableBitmap);
                }
            }
            else
            {
                if (oNoiseBlenderEffect.absolute != absolute)
                {
                    oNoiseBlenderEffect = new NoiseBlenderEffect(octaves - 1, absolute);
                    oNoiseBlenderEffect.aInput = aImageBrush;
                }
                oNoiseBlenderEffect.c1 = c1;
                oNoiseBlenderEffect.c2 = c2;
                oNoiseBlenderEffect.aAmplitude = aAmplitude;
                for (int i = 0; i < octaves; i++)
                {
                    if (tileableX || tileableY)
                    {
                        oPerlinNoise2DClass.tileableNoise(freqXY[i], aOffset[i], absolute, tileableX, tileableY, aWriteableBitmap[i]);
                    }
                    else
                    {
                        oPerlinNoise2DClass.noise(freqXY[i], aOffset[i], absolute, aWriteableBitmap[i]);
                    }
                }
                oPerlinNoise2DClass.currentUIElement.Effect = oNoiseBlenderEffect;
                oWriteableBitmap.Render(oPerlinNoise2DClass.currentUIElement, oTransform);
                oWriteableBitmap.Invalidate();

            }
        }

        /// <summary>
        /// calculates amplitude*cos(x*xyMult.x+y*xyMult.y+sum(|noise|)+shift)
        /// </summary>
        /// <param name="freqXY">frequencies at the x and y axis</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="amplitude">amplitude of the cosine function</param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="tileable">true, if tileable</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void noises(Size[] freqXY, int octaves, int seed, Size[] aOffset, float[] aAmplitude, float amplitude, Point xyMult, float shift, bool tileable, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset.Length != octaves) return;
            if (aAmplitude.Length != octaves) return;
            if (freqXY.Length != octaves) return;
            bool isNotEqualSize = oPerlinNoise2DClass == null ? true : !(new Size(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight)).Equals(oPerlinNoise2DClass.thisSize);
            if (octaves != this.octavesCos || isNotEqualSize)
            {
                this.octaves = octaves;
                oNoiseBlenderCosineEffect = new NoiseBlenderCosineEffect(octaves - 1);
                aWriteableBitmap = new WriteableBitmap[octaves];
                aImageBrush = new ImageBrush[octaves];
                for (int i = 0; i < octaves; i++)
                {
                    aWriteableBitmap[i] = new WriteableBitmap(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight);
                    aImageBrush[i] = new ImageBrush();
                    aImageBrush[i].ImageSource = aWriteableBitmap[i];
                }
                oNoiseBlenderCosineEffect.aInput = aImageBrush;
            }


            if (oPerlinNoise2DClass == null)
            {
                this.seed = seed;
                oPerlinNoise2DClass = new PerlinNoise2DClass(seed);
                oPerlinNoise2DClass.customUIElement = iUIElement;
                oPerlinNoise2DClass.oCustomTransform = oTransform;
            }
            else if (this.seed != seed)
            {
                this.seed = seed;
                oPerlinNoise2DClass.reInitialize(seed);
            }
            oNoiseBlenderCosineEffect.c1 = c1;
            oNoiseBlenderCosineEffect.c2 = c2;
            oNoiseBlenderCosineEffect.aAmplitude = aAmplitude;
            oNoiseBlenderCosineEffect.amplitude = 0.5f * amplitude;
            oNoiseBlenderCosineEffect.xyMult = new Point(xyMult.X * oWriteableBitmap.PixelWidth, xyMult.Y * oWriteableBitmap.PixelHeight);
            oNoiseBlenderCosineEffect.shift = shift;
            for (int i = 0; i < octaves; i++)
            {
                if (tileable)
                {
                    oPerlinNoise2DClass.tileableNoise(freqXY[i], aOffset[i], true, aWriteableBitmap[i]);
                }
                else
                {
                    oPerlinNoise2DClass.noise(freqXY[i], aOffset[i], true, aWriteableBitmap[i]);
                }
            }
            oPerlinNoise2DClass.currentUIElement.Effect = oNoiseBlenderCosineEffect;
            oWriteableBitmap.Render(oPerlinNoise2DClass.currentUIElement, oTransform);
            oWriteableBitmap.Invalidate();
        }

        /// <summary>
        /// calculates amplitude*cos(x*xyMult.x+y*xyMult.y+sum(|noise|)+shift)
        /// </summary>
        /// <param name="freqXY">frequencies at the x and y axis</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="amplitude">amplitude of the cosine function</param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="tileableX">true, if tileable at the width</param>
        /// <param name="tileableY">true, if tileable at the height</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void noises(Size[] freqXY, int octaves, int seed, Size[] aOffset, float[] aAmplitude, float amplitude, Point xyMult, float shift, bool tileableX, bool tileableY, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset.Length != octaves) return;
            if (aAmplitude.Length != octaves) return;
            if (freqXY.Length != octaves) return;
            bool isNotEqualSize = oPerlinNoise2DClass == null ? true : !(new Size(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight)).Equals(oPerlinNoise2DClass.thisSize);
            if (octaves != this.octavesCos || isNotEqualSize)
            {
                this.octaves = octaves;
                oNoiseBlenderCosineEffect = new NoiseBlenderCosineEffect(octaves - 1);
                aWriteableBitmap = new WriteableBitmap[octaves];
                aImageBrush = new ImageBrush[octaves];
                for (int i = 0; i < octaves; i++)
                {
                    aWriteableBitmap[i] = new WriteableBitmap(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight);
                    aImageBrush[i] = new ImageBrush();
                    aImageBrush[i].ImageSource = aWriteableBitmap[i];
                }
            }


            if (oPerlinNoise2DClass == null)
            {
                this.seed = seed;
                oPerlinNoise2DClass = new PerlinNoise2DClass(seed);
                oPerlinNoise2DClass.customUIElement = iUIElement;
                oPerlinNoise2DClass.oCustomTransform = oTransform;
            }
            else if (this.seed != seed)
            {
                this.seed = seed;
                oPerlinNoise2DClass.reInitialize(seed);
            }
            oNoiseBlenderCosineEffect.aInput = aImageBrush;
            oNoiseBlenderCosineEffect.c1 = c1;
            oNoiseBlenderCosineEffect.c2 = c2;
            oNoiseBlenderCosineEffect.aAmplitude = aAmplitude;
            oNoiseBlenderCosineEffect.amplitude = 0.5f * amplitude;
            oNoiseBlenderCosineEffect.xyMult = new Point(xyMult.X * oWriteableBitmap.PixelWidth, xyMult.Y * oWriteableBitmap.PixelHeight);
            oNoiseBlenderCosineEffect.shift = shift;
            for (int i = 0; i < octaves; i++)
            {
                if (tileableX || tileableY)
                {
                    oPerlinNoise2DClass.tileableNoise(freqXY[i], aOffset[i], true, tileableX, tileableY, aWriteableBitmap[i]);
                }
                else
                {
                    oPerlinNoise2DClass.noise(freqXY[i], aOffset[i], true, aWriteableBitmap[i]);
                }
            }
            oPerlinNoise2DClass.currentUIElement.Effect = oNoiseBlenderCosineEffect;
            oWriteableBitmap.Render(oPerlinNoise2DClass.currentUIElement, oTransform);
            oWriteableBitmap.Invalidate();
        }

        /// <summary>
        /// calculates sum(noise) or sum(|noise|) with defined amplitudes and frequencies at higher octaves and with a random seed
        /// </summary>
        /// <param name="freqXY">frequency of the first octave</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="tileable">true, if tileable</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void turbulence(Size freqXY, int octaves, Size[] aOffset, float amplitude, float persistence, bool absolute, bool tileable, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            turbulence(freqXY, octaves, (new Random()).Next(), aOffset, amplitude, persistence, absolute, tileable, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates sum(noise) or sum(|noise|) with defined amplitudes and frequencies at higher octaves and with a defined seed
        /// </summary>
        /// <param name="freqXY">frequency of the first octave</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="tileable">true, if tileable</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void turbulence(Size freqXY, int octaves, int seed, Size[] aOffset, float amplitude, float persistence, bool absolute, bool tileable, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset == null) aOffset = new Size[octaves];
            Size[] aFreqXY = new Size[octaves];
            float[] aAmplitudes = new float[octaves];
            double freqWidth = freqXY.Width;
            double freqHeight = freqXY.Height;
            for (int i = 0; i < octaves; i++)
            {
                if (aOffset[i] == null) aOffset[i] = new Size(0, 0);
                aFreqXY[i] = new Size(freqWidth, freqHeight);
                freqWidth *= 2.0;
                freqHeight *= 2.0;
                aAmplitudes[i] = amplitude;
                amplitude *= persistence;
            }
            noises(aFreqXY, octaves, seed, aOffset, aAmplitudes, absolute, tileable, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates sum(noise) or sum(|noise|) with defined amplitudes and frequencies at higher octaves and with a random seed
        /// </summary>
        /// <param name="freqXY">frequency of the first octave</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="tileableX">true, if tileable at the width</param>
        /// <param name="tileableY">true, if tileable at the height</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void turbulence(Size freqXY, int octaves, Size[] aOffset, float amplitude, float persistence, bool absolute, bool tileableX, bool tileableY, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            turbulence(freqXY, octaves, (new Random()).Next(), aOffset, amplitude, persistence, absolute, tileableX, tileableY, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates sum(noise) or sum(|noise|) with defined amplitudes and frequencies at higher octaves and with a defined seed
        /// </summary>
        /// <param name="freqXY">frequency of the first octave</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="tileableX">true, if tileable at the width</param>
        /// <param name="tileableY">true, if tileable at the height</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void turbulence(Size freqXY, int octaves, int seed, Size[] aOffset, float amplitude, float persistence, bool absolute, bool tileableX, bool tileableY, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset == null) aOffset = new Size[octaves];
            Size[] aFreqXY = new Size[octaves];
            float[] aAmplitudes = new float[octaves];
            double freqWidth = freqXY.Width;
            double freqHeight = freqXY.Height;
            for (int i = 0; i < octaves; i++)
            {
                if (aOffset[i] == null) aOffset[i] = new Size(0, 0);
                aFreqXY[i] = new Size(freqWidth, freqHeight);
                freqWidth *= 2.0;
                freqHeight *= 2.0;
                aAmplitudes[i] = amplitude;
                amplitude *= persistence;
            }
            noises(aFreqXY, octaves, seed, aOffset, aAmplitudes, absolute, tileableX, tileableY, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates amplitudeCos*cos(x*xyMult.x+y*xyMult.y+sum(|noise|)+shift) with a random seed
        /// </summary>
        /// <param name="freqXY">frequency of the first octave</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="amplitudeCos"></param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="tileable">true, if tileable</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void turbulence(Size freqXY, int octaves, Size[] aOffset, float amplitude, float persistence, float amplitudeCos, Point xyMult, float shift, bool tileable, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            turbulence(freqXY, octaves, (new Random()).Next(), aOffset, amplitude, persistence, amplitudeCos,xyMult,shift, tileable, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates amplitudeCos*cos(x*xyMult.x+y*xyMult.y+sum(|noise|)+shift) with a defined seed
        /// </summary>
        /// <param name="freqXY">frequency of the first octave</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="amplitudeCos"></param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="tileable">true, if tileable</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void turbulence(Size freqXY, int octaves, int seed, Size[] aOffset, float amplitude, float persistence, float amplitudeCos, Point xyMult, float shift, bool tileable, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset == null) aOffset = new Size[octaves];
            Size[] aFreqXY = new Size[octaves];
            float[] aAmplitudes = new float[octaves];
            double freqWidth = freqXY.Width;
            double freqHeight = freqXY.Height;
            for (int i = 0; i < octaves; i++)
            {
                if (aOffset[i] == null) aOffset[i] = new Size(0, 0);
                aFreqXY[i] = new Size(freqWidth, freqHeight);
                freqWidth *= 2.0;
                freqHeight *= 2.0;
                aAmplitudes[i] = amplitude;
                amplitude *= persistence;
            }
            noises(aFreqXY, octaves, seed, aOffset, aAmplitudes, amplitudeCos,xyMult,shift, tileable, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates amplitudeCos*cos(x*xyMult.x+y*xyMult.y+sum(|noise|)+shift) with a random seed
        /// </summary>
        /// <param name="freqXY">frequency of the first octave</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="amplitudeCos"></param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="tileableX">true, if tileable at the width</param>
        /// <param name="tileableY">true, if tileable at the height</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void turbulence(Size freqXY, int octaves, Size[] aOffset, float amplitude, float persistence, float amplitudeCos, Point xyMult, float shift, bool tileableX, bool tileableY, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            turbulence(freqXY, octaves, (new Random()).Next(), aOffset, amplitude, persistence, amplitudeCos,xyMult,shift, tileableX, tileableY, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates amplitudeCos*cos(x*xyMult.x+y*xyMult.y+sum(|noise|)+shift) with a defined seed
        /// </summary>
        /// <param name="freqXY">frequency of the first octave</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="amplitudeCos"></param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="tileableX">true, if tileable at the width</param>
        /// <param name="tileableY">true, if tileable at the height</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void turbulence(Size freqXY, int octaves, int seed, Size[] aOffset, float amplitude, float persistence, float amplitudeCos, Point xyMult, float shift, bool tileableX, bool tileableY, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset == null) aOffset = new Size[octaves];
            Size[] aFreqXY = new Size[octaves];
            float[] aAmplitudes = new float[octaves];
            double freqWidth = freqXY.Width;
            double freqHeight = freqXY.Height;
            for (int i = 0; i < octaves; i++)
            {
                if (aOffset[i] == null) aOffset[i] = new Size(0, 0);
                aFreqXY[i] = new Size(freqWidth, freqHeight);
                freqWidth *= 2.0;
                freqHeight *= 2.0;
                aAmplitudes[i] = amplitude;
                amplitude *= persistence;
            }
            noises(aFreqXY, octaves, seed, aOffset, aAmplitudes, amplitudeCos,xyMult,shift, tileableX, tileableY, c1, c2, oWriteableBitmap);
        }

        #endregion

        #region improved 2D noise
        
        /// <summary>
        /// calculates sum(noise) or sum(|noise|) with a random seed value
        /// </summary>
        /// <param name="freq">frequencies at the x and y axis</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destintation bitmap</param>
        public void noises2D(Size[] freq, int octaves, Size[] aOffset, float[] aAmplitude, bool absolute, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            noises2D(freq, octaves, (new Random()).Next(), aOffset, aAmplitude, absolute, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates tileable sum(noise) or sum(|noise|) with a random seed value
        /// </summary>
        /// <param name="freq">frequencies at the x and y axis</param>
        /// <param name="tile">if freq=(i/width, j/height), the noises tiles at the borders for i=2..256 and j=2..256</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destintation bitmap</param>
        public void noises2D(Size[] freq, iSize[] tile, int octaves, Size[] aOffset, float[] aAmplitude, bool absolute, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            noises2D(freq, tile, octaves, (new Random()).Next(), aOffset, aAmplitude, absolute, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates amplitude*cos(x*xyMult.x+y*xyMult.y+sum(|noise|)+shift) with a random seed value
        /// </summary>
        /// <param name="freq">frequencies at the x and y axis</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="amplitude">amplitude of the cosine function</param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void noises2D(Size[] freq, int octaves, Size[] aOffset, float[] aAmplitude, float amplitude, Point xyMult, float shift, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            noises2D(freq, octaves, (new Random()).Next(), aOffset, aAmplitude, amplitude, xyMult, shift, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates amplitude*cos(x*xyMult.x+y*xyMult.y+sum(|noise|)+shift) with a random seed value and tileable noise
        /// </summary>
        /// <param name="freq">frequencies at the x and y axis</param>
        /// <param name="tile">if freq=(i/width, j/height), the noises tiles at the borders for i=2..256 and j=2..256</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="amplitude">amplitude of the cosine function</param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void noises2D(Size[] freq, iSize[] tile, int octaves, Size[] aOffset, float[] aAmplitude, float amplitude, Point xyMult, float shift, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            noises2D(freq, tile, octaves, (new Random()).Next(), aOffset, aAmplitude, amplitude, xyMult, shift, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates sum(noise) or sum(|noise|) with a defined seed value
        /// </summary>
        /// <param name="freq">frequencies at the x and y axis</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destintation bitmap</param>
        public void noises2D(Size[] freq, int octaves, int seed, Size[] aOffset, float[] aAmplitude, bool absolute, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset.Length != octaves) return;
            if (aAmplitude.Length != octaves) return;
            if (freq.Length != octaves) return;
            bool isNotEqualSize= oNoise2D==null?true:!(new Size(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight)).Equals(oNoise2D.thisSize);
            if (octaves != this.octaves || isNotEqualSize)
            {
                this.octaves = octaves;
                if (octaves > 1 )
                {
                    oNoiseBlenderEffect = new NoiseBlenderEffect(octaves - 1, absolute);
                    aWriteableBitmap = new WriteableBitmap[octaves];
                    aImageBrush = new ImageBrush[octaves];
                    for (int i = 0; i < octaves; i++)
                    {
                        aWriteableBitmap[i] = new WriteableBitmap(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight);
                        aImageBrush[i] = new ImageBrush();
                        aImageBrush[i].ImageSource = aWriteableBitmap[i];
                    }
                    oNoiseBlenderEffect.aInput = aImageBrush;
                    }
                else oNoiseBlenderEffect = null;

            }

            if (oNoise2D == null)
            {
                this.seed = seed;
                oNoise2D = new Noise2D(seed);
                oNoise2D.customUIElement = iUIElement;
                oNoise2D.oCustomTransform = oTransform;
            }
            else if (this.seed != seed)
            {
                this.seed = seed;
                oNoise2D.reInitialize(seed);
            }
            if (octaves == 1)
            {
                oNoise2D.singleNoise(freq[0], aOffset[0], aAmplitude[0], absolute, c1, c2, oWriteableBitmap);
            }
            else
            {
                if (oNoiseBlenderEffect.absolute != absolute)
                {
                    oNoiseBlenderEffect = new NoiseBlenderEffect(octaves - 1, absolute);
                    oNoiseBlenderEffect.aInput = aImageBrush;
                    
                }
                oNoiseBlenderEffect.c1 = c1;
                oNoiseBlenderEffect.c2 = c2;
                oNoiseBlenderEffect.aAmplitude = aAmplitude;
                for (int i = 0; i < octaves; i++)
                {
                    oNoise2D.noise(freq[i], aOffset[i], absolute, aWriteableBitmap[i]);
                }
                oNoise2D.currentUIElement.Effect = oNoiseBlenderEffect;
                oWriteableBitmap.Render(oNoise2D.currentUIElement, oTransform);
                oWriteableBitmap.Invalidate();

            }
        }

        /// <summary>
        /// calculates tileable sum(noise) or sum(|noise|) with a defined seed value
        /// </summary>
        /// <param name="freq">frequencies at the x and y axis</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="tile">if freq=(i/width, j/height), the noises tiles at the borders for i=2..256 and j=2..256</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destintation bitmap</param>
        public void noises2D(Size[] freq, iSize[] tile, int octaves, int seed, Size[] aOffset, float[] aAmplitude, bool absolute, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset.Length != octaves) return;
            if (aAmplitude.Length != octaves) return;
            if (freq.Length != octaves) return;
            bool isNotEqualSize = oNoise2D == null ? true : !(new Size(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight)).Equals(oNoise2D.thisSize);
            if (octaves != this.octaves || isNotEqualSize)
            {
                this.octaves = octaves;
                if (octaves > 1)
                {
                    oNoiseBlenderEffect = new NoiseBlenderEffect(octaves - 1, absolute);
                    aWriteableBitmap = new WriteableBitmap[octaves];
                    aImageBrush = new ImageBrush[octaves];
                    for (int i = 0; i < octaves; i++)
                    {
                        aWriteableBitmap[i] = new WriteableBitmap(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight);
                        aImageBrush[i] = new ImageBrush();
                        aImageBrush[i].ImageSource = aWriteableBitmap[i];
                    }
                    oNoiseBlenderEffect.aInput = aImageBrush;
                }
                else oNoiseBlenderEffect = null;

            }

            if (oNoise2D == null)
            {
                this.seed = seed;
                oNoise2D = new Noise2D(seed);
                oNoise2D.customUIElement = iUIElement;
                oNoise2D.oCustomTransform = oTransform;
            }
            else if (this.seed != seed)
            {
                this.seed = seed;
                oNoise2D.reInitialize(seed);
            }
            if (octaves == 1)
            {
                oNoise2D.singleNoise(freq[0], tile[0], aOffset[0], aAmplitude[0], absolute, c1, c2, oWriteableBitmap);
            }
            else
            {
                if (oNoiseBlenderEffect.absolute != absolute)
                {
                    oNoiseBlenderEffect = new NoiseBlenderEffect(octaves - 1, absolute);
                    oNoiseBlenderEffect.aInput = aImageBrush;
                }
                oNoiseBlenderEffect.c1 = c1;
                oNoiseBlenderEffect.c2 = c2;
                oNoiseBlenderEffect.aAmplitude = aAmplitude;
                for (int i = 0; i < octaves; i++)
                {
                        oNoise2D.noise(freq[i], tile[i], aOffset[i], absolute, aWriteableBitmap[i]);
                }
                oNoise2D.currentUIElement.Effect = oNoiseBlenderEffect;
                oWriteableBitmap.Render(oNoise2D.currentUIElement, oTransform);
                oWriteableBitmap.Invalidate();

            }
        }

        /// <summary>
        /// calculates amplitude*cos(x*xyMult.x+y*xyMult.y+sum(|noise|)+shift) with a defined seed value
        /// </summary>
        /// <param name="freq">frequencies at the x and y axis</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="amplitude">amplitude of the cosine function</param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void noises2D(Size[] freq, int octaves, int seed, Size[] aOffset, float[] aAmplitude, float amplitude, Point xyMult, float shift, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset.Length != octaves) return;
            if (aAmplitude.Length != octaves) return;
            if (freq.Length != octaves) return;
            bool isNotEqualSize = oNoise2D == null ? true : !(new Size(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight)).Equals(oNoise2D.thisSize);
            if (octaves != this.octavesCos || isNotEqualSize)
            {
                this.octaves = octaves;
                oNoiseBlenderCosineEffect = new NoiseBlenderCosineEffect(octaves - 1);
                aWriteableBitmap = new WriteableBitmap[octaves];
                aImageBrush = new ImageBrush[octaves];
                for (int i = 0; i < octaves; i++)
                {
                    aWriteableBitmap[i] = new WriteableBitmap(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight);
                    aImageBrush[i] = new ImageBrush();
                    aImageBrush[i].ImageSource = aWriteableBitmap[i];
                }
                oNoiseBlenderCosineEffect.aInput = aImageBrush;
            }


            if (oNoise2D == null)
            {
                this.seed = seed;
                oNoise2D = new Noise2D(seed);
                oNoise2D.customUIElement = iUIElement;
                oNoise2D.oCustomTransform = oTransform;
            }
            else if (this.seed != seed)
            {
                this.seed = seed;
                oNoise2D.reInitialize(seed);
            }
            oNoiseBlenderCosineEffect.c1 = c1;
            oNoiseBlenderCosineEffect.c2 = c2;
            oNoiseBlenderCosineEffect.aAmplitude = aAmplitude;
            oNoiseBlenderCosineEffect.amplitude = 0.5f * amplitude;
            oNoiseBlenderCosineEffect.xyMult = new Point(xyMult.X * oWriteableBitmap.PixelWidth, xyMult.Y * oWriteableBitmap.PixelHeight);
            oNoiseBlenderCosineEffect.shift = shift;
            for (int i = 0; i < octaves; i++)
            {
                oNoise2D.noise(freq[i], aOffset[i], true, aWriteableBitmap[i]);
            }
            oNoise2D.currentUIElement.Effect = oNoiseBlenderCosineEffect;
            oWriteableBitmap.Render(oNoise2D.currentUIElement, oTransform);
            oWriteableBitmap.Invalidate();
        }

        /// <summary>
        /// calculates amplitude*cos(x*xyMult.x+y*xyMult.y+sum(|noise|)+shift) with a defined seed value and tileable noise
        /// </summary>
        /// <param name="freq">frequencies at the x and y axis</param>
        /// <param name="tile">if freq=(i/width, j/height), the noises tiles at the borders for i=2..256 and j=2..256</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="amplitude">amplitude of the cosine function</param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void noises2D(Size[] freq, iSize[] tile, int octaves, int seed, Size[] aOffset, float[] aAmplitude, float amplitude, Point xyMult, float shift, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset.Length != octaves) return;
            if (aAmplitude.Length != octaves) return;
            if (freq.Length != octaves) return;
            bool isNotEqualSize = oNoise2D == null ? true : !(new Size(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight)).Equals(oNoise2D.thisSize);
            if (octaves != this.octavesCos || isNotEqualSize)
            {
                this.octaves = octaves;
                oNoiseBlenderCosineEffect = new NoiseBlenderCosineEffect(octaves - 1);
                aWriteableBitmap = new WriteableBitmap[octaves];
                aImageBrush = new ImageBrush[octaves];
                for (int i = 0; i < octaves; i++)
                {
                    aWriteableBitmap[i] = new WriteableBitmap(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight);
                    aImageBrush[i] = new ImageBrush();
                    aImageBrush[i].ImageSource = aWriteableBitmap[i];
                }
                oNoiseBlenderCosineEffect.aInput = aImageBrush;
            }


            if (oNoise2D == null)
            {
                this.seed = seed;
                oNoise2D = new Noise2D(seed);
                oNoise2D.customUIElement = iUIElement;
                oNoise2D.oCustomTransform = oTransform;
            }
            else if (this.seed != seed)
            {
                this.seed = seed;
                oNoise2D.reInitialize(seed);
            }
            oNoiseBlenderCosineEffect.c1 = c1;
            oNoiseBlenderCosineEffect.c2 = c2;
            oNoiseBlenderCosineEffect.aAmplitude = aAmplitude;
            oNoiseBlenderCosineEffect.amplitude = 0.5f * amplitude;
            oNoiseBlenderCosineEffect.xyMult = new Point(xyMult.X * oWriteableBitmap.PixelWidth, xyMult.Y * oWriteableBitmap.PixelHeight);
            oNoiseBlenderCosineEffect.shift = shift;
            for (int i = 0; i < octaves; i++)
            {
                oNoise2D.noise(freq[i], tile[i], aOffset[i], true, aWriteableBitmap[i]);
            }
            oNoise2D.currentUIElement.Effect = oNoiseBlenderCosineEffect;
            oWriteableBitmap.Render(oNoise2D.currentUIElement, oTransform);
            oWriteableBitmap.Invalidate();
        }


        /// <summary>
        /// calculates sum(noise) or sum(|noise|) with defined amplitudes and frequencies at higher octaves and with a random seed
        /// </summary>
        /// <param name="freq">frequency of the first octave</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void turbulence2D(Size freq, int octaves, Size[] aOffset, float amplitude, float persistence, bool absolute, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            turbulence2D(freq, octaves, (new Random()).Next(), aOffset, amplitude, persistence, absolute, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates sum(noise) or sum(|noise|) with defined amplitudes and frequencies at higher octaves and with a defined seed
        /// </summary>
        /// <param name="freq">frequency of the first octave</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void turbulence2D(Size freq, int octaves, int seed, Size[] aOffset, float amplitude, float persistence, bool absolute, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset == null) aOffset = new Size[octaves];
            Size[] aFreqXY = new Size[octaves];
            float[] aAmplitudes = new float[octaves];
            double freqWidth = freq.Width;
            double freqHeight = freq.Height;
            for (int i = 0; i < octaves; i++)
            {
                if (aOffset[i] == null) aOffset[i] = new Size(0, 0);
                aFreqXY[i] = new Size(freqWidth, freqHeight);
                freqWidth *= 2.0;
                freqHeight *= 2.0;
                aAmplitudes[i] = amplitude;
                amplitude *= persistence;
            }
            noises2D(aFreqXY, octaves, seed, aOffset, aAmplitudes, absolute, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates tileable sum(noise) or sum(|noise|) with defined amplitudes and frequencies at higher octaves and with a random seed
        /// </summary>
        /// <param name="freq">frequency of the first octave</param>
        /// <param name="tile">tile of the first octave; if freq=(i/width, j/height), the noise tiles at the borders for i=2..256 and j=2..256</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void turbulence2D(Size freq, iSize tile, int octaves, Size[] aOffset, float amplitude, float persistence, bool absolute, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            turbulence2D(freq, tile, octaves, (new Random()).Next(), aOffset, amplitude, persistence, absolute, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates tileable sum(noise) or sum(|noise|) with defined amplitudes and frequencies at higher octaves and with a defined seed
        /// </summary>
        /// <param name="freq">frequency of the first octave</param>
        /// <param name="tile">tile of the first octave; if freq=(i/width, j/height), the noise tiles at the borders for i=2..256 and j=2..256</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void turbulence2D(Size freq, iSize tile, int octaves, int seed, Size[] aOffset, float amplitude, float persistence, bool absolute, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset == null) aOffset = new Size[octaves];
            Size[] aFreqXY = new Size[octaves];
            float[] aAmplitudes = new float[octaves];
            double freqWidth = freq.Width;
            double freqHeight = freq.Height;
            iSize[] aTile = new iSize[octaves];
            iSize oTile = (tile != null) ? new iSize(tile.Width, tile.Height) : new iSize(256, 256);
            for (int i = 0; i < octaves; i++)
            {
                if (aOffset[i] == null) aOffset[i] = new Size(0, 0);
                aFreqXY[i] = new Size(freqWidth, freqHeight);
                aTile[i] = new iSize(oTile.Width, oTile.Height);
                oTile.Width *= 2;
                oTile.Height *= 2;
                freqWidth *= 2.0;
                freqHeight *= 2.0;
                aAmplitudes[i] = amplitude;
                amplitude *= persistence;
            }
            noises2D(aFreqXY, aTile, octaves, seed, aOffset, aAmplitudes, absolute, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates amplitudeCos*cos(x*xyMult.x+y*xyMult.y+sum(|noise|)+shift) with a random seed
        /// </summary>
        /// <param name="freq">frequency of the first octave</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="amplitudeCos"></param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void turbulence2D(Size freq, int octaves, Size[] aOffset, float amplitude, float persistence, float amplitudeCos, Point xyMult, float shift, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            turbulence2D(freq, octaves, (new Random()).Next(), aOffset, amplitude, persistence, amplitudeCos, xyMult, shift, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates amplitudeCos*cos(x*xyMult.x+y*xyMult.y+sum(|noise|)+shift) with a defined seed
        /// </summary>
        /// <param name="freq">frequency of the first octave</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="amplitudeCos"></param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void turbulence2D(Size freq, int octaves, int seed, Size[] aOffset, float amplitude, float persistence, float amplitudeCos, Point xyMult, float shift, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset == null) aOffset = new Size[octaves];
            Size[] aFreqXY = new Size[octaves];
            float[] aAmplitudes = new float[octaves];
            double freqWidth = freq.Width;
            double freqHeight = freq.Height;
            for (int i = 0; i < octaves; i++)
            {
                if (aOffset[i] == null) aOffset[i] = new Size(0, 0);
                aFreqXY[i] = new Size(freqWidth, freqHeight);
                freqWidth *= 2.0;
                freqHeight *= 2.0;
                aAmplitudes[i] = amplitude;
                amplitude *= persistence;
            }
            noises2D(aFreqXY, octaves, seed, aOffset, aAmplitudes, amplitudeCos, xyMult, shift, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates amplitudeCos*cos(x*xyMult.x+y*xyMult.y+sum(|noise|)+shift) with a random defined seed and tileable noise
        /// </summary>
        /// <param name="freq">frequency of the first octave</param>
        /// <param name="tile">tile of the first octave; if freq=(i/width, j/height), the noise tiles at the borders for i=2..256 and j=2..256</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="amplitudeCos"></param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void turbulence2D(Size freq, iSize tile, int octaves, Size[] aOffset, float amplitude, float persistence, float amplitudeCos, Point xyMult, float shift, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            turbulence2D(freq, tile, octaves, (new Random()).Next(), aOffset, amplitude, persistence, amplitudeCos, xyMult, shift, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates amplitudeCos*cos(x*xyMult.x+y*xyMult.y+sum(|noise|)+shift) with a defined seed and tileable noise
        /// </summary>
        /// <param name="freq">frequency of the first octave</param>
        /// <param name="tile">tile of the first octave; if freq=(i/width, j/height), the noise tiles at the borders for i=2..256 and j=2..256</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="amplitudeCos"></param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void turbulence2D(Size freq, iSize tile, int octaves, int seed, Size[] aOffset, float amplitude, float persistence, float amplitudeCos, Point xyMult, float shift, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset == null) aOffset = new Size[octaves];
            Size[] aFreqXY = new Size[octaves];
            float[] aAmplitudes = new float[octaves];
            double freqWidth = freq.Width;
            double freqHeight = freq.Height;
            iSize[] aTile = new iSize[octaves];
            iSize oTile = (tile != null) ? new iSize(tile.Width, tile.Height) : new iSize(256, 256);
            for (int i = 0; i < octaves; i++)
            {
                if (aOffset[i] == null) aOffset[i] = new Size(0, 0);
                aFreqXY[i] = new Size(freqWidth, freqHeight);
                aTile[i] = new iSize(oTile.Width, oTile.Height);
                oTile.Width *= 2;
                oTile.Height *= 2;
                freqWidth *= 2.0;
                freqHeight *= 2.0;
                aAmplitudes[i] = amplitude;
                amplitude *= persistence;
            }
            noises2D(aFreqXY, aTile, octaves, seed, aOffset, aAmplitudes, amplitudeCos, xyMult, shift, c1, c2, oWriteableBitmap);
        }
        #endregion

        #region simplex 2D noise
        
        /// <summary>
        /// calculates sum(noise) or sum(|noise|) with a random seed value
        /// </summary>
        /// <param name="freqXY">frequencies at the x and y axis</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destintation bitmap</param>
        public void simplexNoises(Size[] freqXY, int octaves, Size[] aOffset, float[] aAmplitude, bool absolute, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            simplexNoises(freqXY, octaves, (new Random()).Next(), aOffset, aAmplitude, absolute, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates amplitude*cos(x*xyMult.x+y*xyMult.y+sum(|noise|)+shift) with a random seed value
        /// </summary>
        /// <param name="freqXY">frequencies at the x and y axis</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="amplitude">amplitude of the cosine function</param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void simplexNoises(Size[] freqXY, int octaves, Size[] aOffset, float[] aAmplitude, float amplitude, Point xyMult, float shift, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            simplexNoises(freqXY, octaves, (new Random()).Next(), aOffset, aAmplitude, amplitude, xyMult, shift, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates sum(noise) or sum(|noise|) with a defined seed value
        /// </summary>
        /// <param name="freqXY">frequencies at the x and y axis</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destintation bitmap</param>
        public void simplexNoises(Size[] freqXY, int octaves, int seed, Size[] aOffset, float[] aAmplitude, bool absolute, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset.Length != octaves) return;
            if (aAmplitude.Length != octaves) return;
            if (freqXY.Length != octaves) return;
            bool isNotEqualSize = oSimplex2DClass == null ? true : !(new Size(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight)).Equals(oSimplex2DClass.thisSize);
            if (octaves != this.octaves || isNotEqualSize)
            {
                this.octaves = octaves;
                if (octaves > 1)
                {
                    oNoiseBlenderEffect = new NoiseBlenderEffect(octaves - 1, absolute);
                    aWriteableBitmap = new WriteableBitmap[octaves];
                    aImageBrush = new ImageBrush[octaves];
                    for (int i = 0; i < octaves; i++)
                    {
                        aWriteableBitmap[i] = new WriteableBitmap(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight);
                        aImageBrush[i] = new ImageBrush();
                        aImageBrush[i].ImageSource = aWriteableBitmap[i];
                    }
                    oNoiseBlenderEffect.aInput = aImageBrush;
                }
                else oNoiseBlenderEffect = null;

            }


            if (oSimplex2DClass == null)
            {
                this.seed = seed;
                oSimplex2DClass = new Simplex2DClass(seed);
                oSimplex2DClass.customUIElement = iUIElement;
                oSimplex2DClass.oCustomTransform = oTransform;
            }
            else if (this.seed != seed)
            {
                this.seed = seed;
                oSimplex2DClass.reInitialize(seed);

            }
            if (octaves == 1)
            {
                oSimplex2DClass.singleNoise(freqXY[0], aOffset[0], aAmplitude[0], absolute, c1, c2, oWriteableBitmap);
                
            }
            else
            {
                if (oNoiseBlenderEffect.absolute != absolute)
                {
                    oNoiseBlenderEffect = new NoiseBlenderEffect(octaves - 1, absolute);
                    oNoiseBlenderEffect.aInput = aImageBrush;
                }
                oNoiseBlenderEffect.c1 = c1;
                oNoiseBlenderEffect.c2 = c2;
                oNoiseBlenderEffect.aAmplitude = aAmplitude;
                for (int i = 0; i < octaves; i++)
                {
                    oSimplex2DClass.noise(freqXY[i], aOffset[i], absolute, aWriteableBitmap[i]);
                }
                oSimplex2DClass.currentUIElement.Effect = oNoiseBlenderEffect;
                oWriteableBitmap.Render(oSimplex2DClass.currentUIElement, oTransform);
                oWriteableBitmap.Invalidate();

            }
        }

        /// <summary>
        /// calculates amplitude*cos(x*xyMult.x+y*xyMult.y+sum(|noise|)+shift) with a defined seed value
        /// </summary>
        /// <param name="freqXY">frequencies at the x and y axis</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="amplitude">amplitude of the cosine function</param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void simplexNoises(Size[] freqXY, int octaves, int seed, Size[] aOffset, float[] aAmplitude, float amplitude, Point xyMult, float shift, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset.Length != octaves) return;
            if (aAmplitude.Length != octaves) return;
            if (freqXY.Length != octaves) return;
            bool isNotEqualSize = oSimplex2DClass == null ? true : !(new Size(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight)).Equals(oSimplex2DClass.thisSize);
            if (octaves != this.octavesCos || isNotEqualSize)
            {
                this.octaves = octaves;
                oNoiseBlenderCosineEffect = new NoiseBlenderCosineEffect(octaves - 1);
                aWriteableBitmap = new WriteableBitmap[octaves];
                aImageBrush = new ImageBrush[octaves];
                for (int i = 0; i < octaves; i++)
                {
                    aWriteableBitmap[i] = new WriteableBitmap(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight);
                    aImageBrush[i] = new ImageBrush();
                    aImageBrush[i].ImageSource = aWriteableBitmap[i];
                }
                oNoiseBlenderCosineEffect.aInput = aImageBrush;
            }


            if (oSimplex2DClass == null)
            {
                this.seed = seed;
                oSimplex2DClass = new Simplex2DClass(seed);
                oSimplex2DClass.customUIElement = iUIElement;
                oSimplex2DClass.oCustomTransform = oTransform;
            }
            else if (this.seed != seed)
            {
                this.seed = seed;
                oSimplex2DClass.reInitialize(seed);

            }
            oNoiseBlenderCosineEffect.c1 = c1;
            oNoiseBlenderCosineEffect.c2 = c2;
            oNoiseBlenderCosineEffect.aAmplitude = aAmplitude;
            oNoiseBlenderCosineEffect.amplitude = 0.5f * amplitude;
            oNoiseBlenderCosineEffect.xyMult = new Point(xyMult.X * oWriteableBitmap.PixelWidth, xyMult.Y * oWriteableBitmap.PixelHeight);
            oNoiseBlenderCosineEffect.shift = shift;
            for (int i = 0; i < octaves; i++)
            {
                oSimplex2DClass.noise(freqXY[i], aOffset[i], true, aWriteableBitmap[i]);
            }
            oSimplex2DClass.currentUIElement.Effect = oNoiseBlenderCosineEffect;
            oWriteableBitmap.Render(oSimplex2DClass.currentUIElement, oTransform);
            oWriteableBitmap.Invalidate();
        }

        /// <summary>
        /// calculates sum(noise) or sum(|noise|) with defined amplitudes and frequencies at higher octaves and with a random seed
        /// </summary>
        /// <param name="freqXY">frequency of the first octave</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void simplexTurbulence(Size freqXY, int octaves, Size[] aOffset, float amplitude, float persistence, bool absolute, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            simplexTurbulence(freqXY, octaves, (new Random()).Next(), aOffset, amplitude, persistence, absolute, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates sum(noise) or sum(|noise|) with defined amplitudes and frequencies at higher octaves and with a defined seed
        /// </summary>
        /// <param name="freqXY">frequency of the first octave</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void simplexTurbulence(Size freqXY, int octaves, int seed, Size[] aOffset, float amplitude, float persistence, bool absolute, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset == null) aOffset = new Size[octaves];
            Size[] aFreqXY = new Size[octaves];
            float[] aAmplitudes = new float[octaves];
            double freqWidth = freqXY.Width;
            double freqHeight = freqXY.Height;
            for (int i = 0; i < octaves; i++)
            {
                if (aOffset[i] == null) aOffset[i] = new Size(0, 0);
                aFreqXY[i] = new Size(freqWidth, freqHeight);
                freqWidth *= 2.0;
                freqHeight *= 2.0;
                aAmplitudes[i] = amplitude;
                amplitude *= persistence;
            }
            simplexNoises(aFreqXY, octaves, seed, aOffset, aAmplitudes, absolute, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates amplitudeCos*cos(x*xyMult.x+y*xyMult.y+sum(|noise|)+shift) with a random seed
        /// </summary>
        /// <param name="freqXY">frequency of the first octave</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="amplitudeCos"></param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void simplexTurbulence(Size freqXY, int octaves, Size[] aOffset, float amplitude, float persistence, float amplitudeCos, Point xyMult, float shift, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            simplexTurbulence(freqXY, octaves, (new Random()).Next(), aOffset, amplitude, persistence, amplitudeCos, xyMult, shift, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates amplitudeCos*cos(x*xyMult.x+y*xyMult.y+sum(|noise|)+shift) with a defined seed
        /// </summary>
        /// <param name="freqXY">frequency of the first octave</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="amplitudeCos"></param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void simplexTurbulence(Size freqXY, int octaves, int seed, Size[] aOffset, float amplitude, float persistence, float amplitudeCos, Point xyMult, float shift, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset == null) aOffset = new Size[octaves];
            Size[] aFreqXY = new Size[octaves];
            float[] aAmplitudes = new float[octaves];
            double freqWidth = freqXY.Width;
            double freqHeight = freqXY.Height;
            for (int i = 0; i < octaves; i++)
            {
                if (aOffset[i] == null) aOffset[i] = new Size(0, 0);
                aFreqXY[i] = new Size(freqWidth, freqHeight);
                freqWidth *= 2.0;
                freqHeight *= 2.0;
                aAmplitudes[i] = amplitude;
                amplitude *= persistence;
            }
            simplexNoises(aFreqXY, octaves, seed, aOffset, aAmplitudes, amplitudeCos, xyMult, shift, c1, c2, oWriteableBitmap);
        }
        #endregion

        #region flat 3D improved noise

        /// <summary>
        /// calculates flat 3D improved perlin noise sum(noise) or sum(|noise|) with a random seed
        /// </summary>
        /// <param name="freqXY">frequencies at the x and y axis</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aZOffset">z values</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination color</param>
        public void improvedNoises(Size[] freqXY, int octaves, Size[] aOffset, float[] aZOffset, float[] aAmplitude, bool absolute, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            improvedNoises(freqXY, octaves, (new Random()).Next(), aOffset, aZOffset, aAmplitude, absolute, c1, c2, oWriteableBitmap);
        }
        
        /// <summary>
        /// calculates flat 3D improved perlin noise amplitude*cos(x*multXY.x+y*multXY.y+sum(|noise|)+shift) with a defined seed
        /// </summary>
        /// <param name="freqXY">frequencies at the x and y axis</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aZOffset">z values</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="amplitude"></param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination color</param>
        public void improvedNoises(Size[] freqXY, int octaves, Size[] aOffset, float[] aZOffset, float[] aAmplitude, float amplitude, Point xyMult, float shift, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            improvedNoises(freqXY, octaves, (new Random()).Next(), aOffset, aZOffset, aAmplitude, amplitude, xyMult, shift, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates flat 3D improved perlin noise sum(noise) or sum(|noise|) with a defined seed
        /// </summary>
        /// <param name="freqXY">frequencies at the x and y axis</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aZOffset">z values</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination color</param>
        public void improvedNoises(Size[] freqXY, int octaves, int seed, Size[] aOffset, float[] aZOffset, float[] aAmplitude, bool absolute, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset.Length != octaves) return;
            if (aAmplitude.Length != octaves) return;
            if (freqXY.Length != octaves) return;
            if (aZOffset.Length != octaves) return;
            bool isNotEqualSize = oImprovedNoiseClass == null ? true : !(new Size(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight)).Equals(oImprovedNoiseClass.thisSize);
            if (octaves != this.octaves || isNotEqualSize)
            {
                this.octaves = octaves;
                if (octaves > 1)
                {
                    oNoiseBlenderEffect = new NoiseBlenderEffect(octaves - 1, absolute);
                    aWriteableBitmap = new WriteableBitmap[octaves];
                    aImageBrush = new ImageBrush[octaves];
                    for (int i = 0; i < octaves; i++)
                    {
                        aWriteableBitmap[i] = new WriteableBitmap(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight);
                        aImageBrush[i] = new ImageBrush();
                        aImageBrush[i].ImageSource = aWriteableBitmap[i];
                    }
                    oNoiseBlenderEffect.aInput = aImageBrush;
                    
                }
                else oNoiseBlenderEffect = null;
            }


            if (oImprovedNoiseClass == null)
            {
                this.seed = seed;
                oImprovedNoiseClass = new ImprovedNoiseClass(seed);
                oImprovedNoiseClass.customUIElement = iUIElement;
                oImprovedNoiseClass.oCustomTransform = oTransform;
            }
            else if (this.seed != seed)
            {
                this.seed = seed;
                oImprovedNoiseClass.reInitialize(seed);

            }
            if (octaves == 1)
            {
                oImprovedNoiseClass.singleNoise(freqXY[0], aOffset[0], aZOffset[0], aAmplitude[0], absolute, c1, c2, oWriteableBitmap);

            }
            else
            {
                if (oNoiseBlenderEffect.absolute != absolute)
                {
                    oNoiseBlenderEffect = new NoiseBlenderEffect(octaves - 1, absolute);
                    oNoiseBlenderEffect.aInput = aImageBrush;
                }
                oNoiseBlenderEffect.c1 = c1;
                oNoiseBlenderEffect.c2 = c2;
                oNoiseBlenderEffect.aAmplitude = aAmplitude;
                for (int i = 0; i < octaves; i++)
                {
                    oImprovedNoiseClass.noise(freqXY[i], aOffset[i], aZOffset[i], absolute, aWriteableBitmap[i]);
                }
                oImprovedNoiseClass.currentUIElement.Effect = oNoiseBlenderEffect;
                oWriteableBitmap.Render(oImprovedNoiseClass.currentUIElement, oTransform);
                oWriteableBitmap.Invalidate();

            }
        }

        /// <summary>
        /// calculates flat 3D improved perlin noise amplitude*cos(x*multXY.x+y*multXY.y+sum(|noise|)+shift) with a defined seed
        /// </summary>
        /// <param name="freqXY">frequencies at the x and y axis</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aZOffset">z values</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="amplitude"></param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination color</param>
        public void improvedNoises(Size[] freqXY, int octaves, int seed, Size[] aOffset, float[] aZOffset, float[] aAmplitude, float amplitude, Point xyMult, float shift, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset.Length != octaves) return;
            if (aAmplitude.Length != octaves) return;
            if (freqXY.Length != octaves) return;
            if (aZOffset.Length != octaves) return;
            bool isNotEqualSize = oImprovedNoiseClass == null ? true : !(new Size(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight)).Equals(oImprovedNoiseClass.thisSize);
            if (octaves != this.octavesCos || isNotEqualSize)
            {
                this.octaves = octaves;
                oNoiseBlenderCosineEffect = new NoiseBlenderCosineEffect(octaves - 1);
                aWriteableBitmap = new WriteableBitmap[octaves];
                aImageBrush = new ImageBrush[octaves];
                for (int i = 0; i < octaves; i++)
                {
                    aWriteableBitmap[i] = new WriteableBitmap(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight);
                    aImageBrush[i] = new ImageBrush();
                    aImageBrush[i].ImageSource = aWriteableBitmap[i];
                }
                oNoiseBlenderCosineEffect.aInput = aImageBrush;
            }


            if (oImprovedNoiseClass == null)
            {
                this.seed = seed;
                oImprovedNoiseClass = new ImprovedNoiseClass(seed);
                oImprovedNoiseClass.customUIElement = iUIElement;
                oImprovedNoiseClass.oCustomTransform = oTransform;
            }
            else if (this.seed != seed)
            {
                this.seed = seed;
                oImprovedNoiseClass.reInitialize(seed);

            }
            oNoiseBlenderCosineEffect.c1 = c1;
            oNoiseBlenderCosineEffect.c2 = c2;
            oNoiseBlenderCosineEffect.aAmplitude = aAmplitude;
            oNoiseBlenderCosineEffect.amplitude = 0.5f * amplitude;
            oNoiseBlenderCosineEffect.xyMult = new Point(xyMult.X * oWriteableBitmap.PixelWidth, xyMult.Y * oWriteableBitmap.PixelHeight);
            oNoiseBlenderCosineEffect.shift = shift;
            for (int i = 0; i < octaves; i++)
            {
                oImprovedNoiseClass.noise(freqXY[i], aOffset[i], aZOffset[i], true, aWriteableBitmap[i]);
            }
            oImprovedNoiseClass.currentUIElement.Effect = oNoiseBlenderCosineEffect;
            oWriteableBitmap.Render(oImprovedNoiseClass.currentUIElement, oTransform);
            oWriteableBitmap.Invalidate();


        }

        /// <summary>
        /// calculates flat, tileable 3D improved perlin noise sum(noise) or sum(|noise|) with a random seed
        /// </summary>
        /// <param name="freqXY">frequencies at the x and y axis</param>
        /// <param name="aTile">if freq=(i/width, j/height), the noises tiles at the borders for i=2..256 and j=2..256</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aZOffset">z values</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination color</param>
        public void improvedNoises(Size[] freqXY, iSize[] aTile, int octaves, Size[] aOffset, float[] aZOffset, float[] aAmplitude, bool absolute, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            improvedNoises(freqXY, aTile, octaves, (new Random()).Next(), aOffset, aZOffset, aAmplitude, absolute, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates flat, tileable 3D improved perlin noise amplitude*cos(x*multXY.x+y*multXY.y+sum(|noise|)+shift) with a defined seed
        /// </summary>
        /// <param name="freqXY">frequencies at the x and y axis</param>
        /// <param name="aTile">if freq=(i/width, j/height), the noises tiles at the borders for i=2..256 and j=2..256</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aZOffset">z values</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="amplitude"></param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination color</param>
        public void improvedNoises(Size[] freqXY, iSize[] aTile, int octaves, Size[] aOffset, float[] aZOffset, float[] aAmplitude, float amplitude, Point xyMult, float shift, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            improvedNoises(freqXY, aTile, octaves, (new Random()).Next(), aOffset, aZOffset, aAmplitude, amplitude, xyMult, shift, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates flat, tileable 3D improved perlin noise sum(noise) or sum(|noise|) with a defined seed
        /// </summary>
        /// <param name="freqXY">frequencies at the x and y axis</param>
        /// <param name="aTile">if freq=(i/width, j/height), the noises tiles at the borders for i=2..256 and j=2..256</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aZOffset">z values</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination color</param>
        public void improvedNoises(Size[] freqXY, iSize[] aTile, int octaves, int seed, Size[] aOffset, float[] aZOffset, float[] aAmplitude, bool absolute, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset.Length != octaves) return;
            if (aAmplitude.Length != octaves) return;
            if (freqXY.Length != octaves) return;
            if (aZOffset.Length != octaves) return;
            if (aTile.Length != octaves) return;
            bool isNotEqualSize = oImprovedNoiseClass == null ? true : !(new Size(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight)).Equals(oImprovedNoiseClass.thisSize);
            if (octaves != this.octaves || isNotEqualSize)
            {
                this.octaves = octaves;
                if (octaves > 1)
                {
                    oNoiseBlenderEffect = new NoiseBlenderEffect(octaves - 1, absolute);
                    aWriteableBitmap = new WriteableBitmap[octaves];
                    aImageBrush = new ImageBrush[octaves];
                    for (int i = 0; i < octaves; i++)
                    {
                        aWriteableBitmap[i] = new WriteableBitmap(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight);
                        aImageBrush[i] = new ImageBrush();
                        aImageBrush[i].ImageSource = aWriteableBitmap[i];
                    }
                    oNoiseBlenderEffect.aInput = aImageBrush;

                }
                else oNoiseBlenderEffect = null;
            }


            if (oImprovedNoiseClass == null)
            {
                this.seed = seed;
                oImprovedNoiseClass = new ImprovedNoiseClass(seed);
                oImprovedNoiseClass.customUIElement = iUIElement;
                oImprovedNoiseClass.oCustomTransform = oTransform;
            }
            else if (this.seed != seed)
            {
                this.seed = seed;
                oImprovedNoiseClass.reInitialize(seed);

            }
            if (octaves == 1)
            {
                oImprovedNoiseClass.singleNoise(freqXY[0], aTile[0], aOffset[0], aZOffset[0], aAmplitude[0], absolute, c1, c2, oWriteableBitmap);

            }
            else
            {
                if (oNoiseBlenderEffect.absolute != absolute)
                {
                    oNoiseBlenderEffect = new NoiseBlenderEffect(octaves - 1, absolute);
                    oNoiseBlenderEffect.aInput = aImageBrush;
                }
                oNoiseBlenderEffect.c1 = c1;
                oNoiseBlenderEffect.c2 = c2;
                oNoiseBlenderEffect.aAmplitude = aAmplitude;
                for (int i = 0; i < octaves; i++)
                {
                    oImprovedNoiseClass.noise(freqXY[i], aTile[i], aOffset[i], aZOffset[i], absolute, aWriteableBitmap[i]);
                }
                oImprovedNoiseClass.currentUIElement.Effect = oNoiseBlenderEffect;
                oWriteableBitmap.Render(oImprovedNoiseClass.currentUIElement, oTransform);
                oWriteableBitmap.Invalidate();

            }
        }

        /// <summary>
        /// calculates flat, tileable 3D improved perlin noise amplitude*cos(x*multXY.x+y*multXY.y+sum(|noise|)+shift) with a defined seed
        /// </summary>
        /// <param name="freqXY">frequencies at the x and y axis</param>
        /// <param name="aTile">if freq=(i/width, j/height), the noises tiles at the borders for i=2..256 and j=2..256</param>
        /// <param name="octaves">number of noises to be mixed</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">upper left corners of the noise planes</param>
        /// <param name="aZOffset">z values</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="amplitude"></param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination color</param>
        public void improvedNoises(Size[] freqXY, iSize[] aTile, int octaves, int seed, Size[] aOffset, float[] aZOffset, float[] aAmplitude, float amplitude, Point xyMult, float shift, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset.Length != octaves) return;
            if (aAmplitude.Length != octaves) return;
            if (freqXY.Length != octaves) return;
            if (aZOffset.Length != octaves) return;
            if (aTile.Length != octaves) return;
            bool isNotEqualSize = oImprovedNoiseClass == null ? true : !(new Size(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight)).Equals(oImprovedNoiseClass.thisSize);
            if (octaves != this.octavesCos || isNotEqualSize)
            {
                this.octaves = octaves;
                oNoiseBlenderCosineEffect = new NoiseBlenderCosineEffect(octaves - 1);
                aWriteableBitmap = new WriteableBitmap[octaves];
                aImageBrush = new ImageBrush[octaves];
                for (int i = 0; i < octaves; i++)
                {
                    aWriteableBitmap[i] = new WriteableBitmap(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight);
                    aImageBrush[i] = new ImageBrush();
                    aImageBrush[i].ImageSource = aWriteableBitmap[i];
                }
                oNoiseBlenderCosineEffect.aInput = aImageBrush;
            }


            if (oImprovedNoiseClass == null)
            {
                this.seed = seed;
                oImprovedNoiseClass = new ImprovedNoiseClass(seed);
                oImprovedNoiseClass.customUIElement = iUIElement;
                oImprovedNoiseClass.oCustomTransform = oTransform;
            }
            else if (this.seed != seed)
            {
                this.seed = seed;
                oImprovedNoiseClass.reInitialize(seed);

            }
            oNoiseBlenderCosineEffect.c1 = c1;
            oNoiseBlenderCosineEffect.c2 = c2;
            oNoiseBlenderCosineEffect.aAmplitude = aAmplitude;
            oNoiseBlenderCosineEffect.amplitude = 0.5f * amplitude;
            oNoiseBlenderCosineEffect.xyMult = new Point(xyMult.X * oWriteableBitmap.PixelWidth, xyMult.Y * oWriteableBitmap.PixelHeight);
            oNoiseBlenderCosineEffect.shift = shift;
            for (int i = 0; i < octaves; i++)
            {
                oImprovedNoiseClass.noise(freqXY[i], aTile[i], aOffset[i], aZOffset[i], true, aWriteableBitmap[i]);
            }
            oImprovedNoiseClass.currentUIElement.Effect = oNoiseBlenderCosineEffect;
            oWriteableBitmap.Render(oImprovedNoiseClass.currentUIElement, oTransform);
            oWriteableBitmap.Invalidate();
        }

        /// <summary>
        /// calculates flat improved 3D Perlin noise sum(noise) or sum(|noise|) with defined amplitudes and frequencies at higher octaves and with a random seed
        /// </summary>
        /// <param name="freqXY">frequency of the first octave</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="aZOffset">z values, can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void improvedTurbulence(Size freqXY, int octaves, Size[] aOffset, float[] aZOffset,float amplitude, float persistence, bool absolute, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            improvedTurbulence(freqXY, octaves, (new Random()).Next(), aOffset, aZOffset, amplitude, persistence, absolute, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates flat improved 3D Perlin noise sum(noise) or sum(|noise|) with defined amplitudes and frequencies at higher octaves and with a defined seed
        /// </summary>
        /// <param name="freqXY">frequency of the first octave</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="aZOffset">z values, can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void improvedTurbulence(Size freqXY, int octaves, int seed, Size[] aOffset, float[] aZOffset, float amplitude, float persistence, bool absolute, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset == null) aOffset = new Size[octaves];
            if (aZOffset == null) aZOffset = new float[octaves];

            Size[] aFreqXY = new Size[octaves];
            float[] aAmplitudes = new float[octaves];
            double freqWidth = freqXY.Width;
            double freqHeight = freqXY.Height;
            for (int i = 0; i < octaves; i++)
            {
                if (aOffset[i] == null) aOffset[i] = new Size(0, 0);
                aFreqXY[i] = new Size(freqWidth, freqHeight);
                freqWidth *= 2.0;
                freqHeight *= 2.0;
                aAmplitudes[i] = amplitude;
                amplitude *= persistence;
            }
            improvedNoises(aFreqXY, octaves, seed, aOffset, aZOffset, aAmplitudes, absolute, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates flat improved 3D Perlin noise with amplitudeCos*cos(x*xyMult.x+y*xyMult.y+sum(|noise|)+shift) with a random seed
        /// </summary>
        /// <param name="freqXY">frequency of the first octave</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="aZOffset">z values, can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="amplitudeCos"></param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void improvedTurbulence(Size freqXY, int octaves, Size[] aOffset, float[] aZOffset, float amplitude, float persistence, float amplitudeCos, Point xyMult, float shift, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            improvedTurbulence(freqXY, octaves, (new Random()).Next(), aOffset, aZOffset, amplitude, persistence, amplitudeCos,xyMult,shift, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates flat improved 3D Perlin noise with amplitudeCos*cos(x*xyMult.x+y*xyMult.y+sum(|noise|)+shift) with a defined seed
        /// </summary>
        /// <param name="freqXY">frequency of the first octave</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="aZOffset">z values, can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="amplitudeCos"></param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void improvedTurbulence(Size freqXY, int octaves, int seed, Size[] aOffset, float[] aZOffset, float amplitude, float persistence, float amplitudeCos, Point xyMult, float shift, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset == null) aOffset = new Size[octaves];
            if (aZOffset == null) aZOffset = new float[octaves];

            Size[] aFreqXY = new Size[octaves];
            float[] aAmplitudes = new float[octaves];
            double freqWidth = freqXY.Width;
            double freqHeight = freqXY.Height;
            for (int i = 0; i < octaves; i++)
            {
                if (aOffset[i] == null) aOffset[i] = new Size(0, 0);
                aFreqXY[i] = new Size(freqWidth, freqHeight);
                freqWidth *= 2.0;
                freqHeight *= 2.0;
                aAmplitudes[i] = amplitude;
                amplitude *= persistence;
            }
            improvedNoises(aFreqXY, octaves, seed, aOffset, aZOffset, aAmplitudes, amplitudeCos,xyMult,shift, c1, c2, oWriteableBitmap);
        }

        
        /// <summary>
        /// calculates flat improved tileable 3D Perlin noise sum(noise) or sum(|noise|) with defined amplitudes and frequencies at higher octaves and with a random seed
        /// </summary>
        /// <param name="freqXY">frequency of the first octave</param>
        /// <param name="tile">tile of the first octave; if freq=(i/width, j/height), the noise tiles at the borders for i=2..256 and j=2..256</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="aZOffset">z values, can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void improvedTurbulence(Size freqXY, iSize tile, int octaves, Size[] aOffset, float[] aZOffset, float amplitude, float persistence, bool absolute, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            improvedTurbulence(freqXY, tile, octaves, (new Random()).Next(), aOffset, aZOffset, amplitude, persistence, absolute, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates flat improved tileable 3D Perlin noise sum(noise) or sum(|noise|) with defined amplitudes and frequencies at higher octaves and with a defined seed
        /// </summary>
        /// <param name="freqXY">frequency of the first octave</param>
        /// <param name="tile">tile of the first octave; if freq=(i/width, j/height), the noise tiles at the borders for i=2..256 and j=2..256</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="aZOffset">z values, can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void improvedTurbulence(Size freqXY, iSize tile, int octaves, int seed, Size[] aOffset, float[] aZOffset, float amplitude, float persistence, bool absolute, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset == null) aOffset = new Size[octaves];
            if (aZOffset == null) aZOffset = new float[octaves];

            Size[] aFreqXY = new Size[octaves];
            float[] aAmplitudes = new float[octaves];
            double freqWidth = freqXY.Width;
            double freqHeight = freqXY.Height;
            iSize[] aTile = new iSize[octaves];
            iSize oTile = (tile != null) ? new iSize(tile.Width, tile.Height) : new iSize(256, 256);
            for (int i = 0; i < octaves; i++)
            {
                if (aOffset[i] == null) aOffset[i] = new Size(0, 0);
                aFreqXY[i] = new Size(freqWidth, freqHeight);
                aTile[i] = new iSize(oTile.Width, oTile.Height);
                oTile.Width *= 2;
                oTile.Height *= 2;
                freqWidth *= 2.0;
                freqHeight *= 2.0;
                aAmplitudes[i] = amplitude;
                amplitude *= persistence;
            }
            improvedNoises(aFreqXY, aTile, octaves, seed, aOffset, aZOffset, aAmplitudes, absolute, c1, c2, oWriteableBitmap);
        }

        
        
        /// <summary>
        /// calculates flat improved tileable 3D Perlin noise with amplitudeCos*cos(x*xyMult.x+y*xyMult.y+sum(|noise|)+shift) with a random seed
        /// </summary>
        /// <param name="freqXY">frequency of the first octave</param>
        /// <param name="tile">tile of the first octave; if freq=(i/width, j/height), the noise tiles at the borders for i=2..256 and j=2..256</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="aZOffset">z values, can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="amplitudeCos"></param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void improvedTurbulence(Size freqXY, iSize tile, int octaves, Size[] aOffset, float[] aZOffset, float amplitude, float persistence, float amplitudeCos, Point xyMult, float shift, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            improvedTurbulence(freqXY, tile, octaves, (new Random()).Next(), aOffset, aZOffset, amplitude, persistence, amplitudeCos, xyMult, shift, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates flat improved tileable 3D Perlin noise with amplitudeCos*cos(x*xyMult.x+y*xyMult.y+sum(|noise|)+shift) with a defined seed
        /// </summary>
        /// <param name="freqXY">frequency of the first octave</param>
        /// <param name="tile">tile of the first octave; if freq=(i/width, j/height), the noise tiles at the borders for i=2..256 and j=2..256</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">upper, left corners of the noise planes; can be null or should match the octaves count</param>
        /// <param name="aZOffset">z values, can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="amplitudeCos"></param>
        /// <param name="xyMult"></param>
        /// <param name="shift"></param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void improvedTurbulence(Size freqXY, iSize tile, int octaves, int seed, Size[] aOffset, float[] aZOffset, float amplitude, float persistence, float amplitudeCos, Point xyMult, float shift, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset == null) aOffset = new Size[octaves];
            if (aZOffset == null) aZOffset = new float[octaves];

            Size[] aFreqXY = new Size[octaves];
            float[] aAmplitudes = new float[octaves];
            double freqWidth = freqXY.Width;
            double freqHeight = freqXY.Height;
            iSize[] aTile = new iSize[octaves];
            iSize oTile = (tile != null) ? new iSize(tile.Width, tile.Height) : new iSize(256, 256);
            for (int i = 0; i < octaves; i++)
            {
                if (aOffset[i] == null) aOffset[i] = new Size(0, 0);
                aFreqXY[i] = new Size(freqWidth, freqHeight);
                freqWidth *= 2.0;
                freqHeight *= 2.0;
                aTile[i] = new iSize(oTile.Width, oTile.Height);
                oTile.Width *= 2;
                oTile.Height *= 2;
                aAmplitudes[i] = amplitude;
                amplitude *= persistence;
            }
            improvedNoises(aFreqXY, aTile, octaves, seed, aOffset, aZOffset, aAmplitudes, amplitudeCos, xyMult, shift, c1, c2, oWriteableBitmap);
        }
        #endregion
        
        #region 3D improved noise for custom objects
        
        /// <summary>
        /// calculates 3D improved Perlin noise sum(noise) or sum(|noise|) for custom objects with a random seed
        /// </summary>
        /// <param name="xPlane">x values of the object with the format r.gba</param>
        /// <param name="yPlane">y values of the object with the format r.gba</param>
        /// <param name="zPlane">z values of the object with the format r.gba</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="aMult">the factors of each noise octave</param>
        /// <param name="aOffset">offset of each noise octave</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void improvedNoises(ImageBrush xPlane, ImageBrush yPlane, ImageBrush zPlane, int octaves, Vector3d[] aMult, Vector3d[] aOffset, float[] aAmplitude, bool absolute, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            improvedNoises(xPlane, yPlane, zPlane, octaves, (new Random()).Next(), aMult, aOffset, aAmplitude, absolute, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates 3D improved Perlin noise sum(noise) or sum(|noise|) for custom objects with a defined seed
        /// </summary>
        /// <param name="xPlane">x values of the object with the format r.gba</param>
        /// <param name="yPlane">y values of the object with the format r.gba</param>
        /// <param name="zPlane">z values of the object with the format r.gba</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aMult">the factors of each noise octave</param>
        /// <param name="aOffset">offset of each noise octave</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void improvedNoises(ImageBrush xPlane, ImageBrush yPlane, ImageBrush zPlane, int octaves, int seed, Vector3d[] aMult, Vector3d[] aOffset, float[] aAmplitude, bool absolute, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset.Length != octaves) return;
            if (aAmplitude.Length != octaves) return;
            if (octaves > 1 && aMult.Length != octaves) return;
            bool isNotEqualSize = oImprovedNoiseClass == null ? true : !(new Size(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight)).Equals(oImprovedNoiseClass.thisSize);
            if (octaves != this.octaves || isNotEqualSize)
            {
                this.octaves = octaves;
                if (octaves > 1)
                {
                    oNoiseBlenderEffect = new NoiseBlenderEffect(octaves - 1, absolute);
                    aWriteableBitmap = new WriteableBitmap[octaves];
                    aImageBrush = new ImageBrush[octaves];
                    for (int i = 0; i < octaves; i++)
                    {
                        aWriteableBitmap[i] = new WriteableBitmap(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight);
                        aImageBrush[i] = new ImageBrush();
                        aImageBrush[i].ImageSource = aWriteableBitmap[i];
                    }
                    oNoiseBlenderEffect.aInput = aImageBrush;
                }
                else oNoiseBlenderEffect = null;

            }


            if (oImprovedNoiseClass == null)
            {
                this.seed = seed;
                oImprovedNoiseClass = new ImprovedNoiseClass(seed);
                oImprovedNoiseClass.customUIElement = iUIElement;
                oImprovedNoiseClass.oCustomTransform = oTransform;
            }
            else if (this.seed != seed)
            {
                this.seed = seed;
                oImprovedNoiseClass.reInitialize(seed);

            }
            if (octaves == 1)
            {
                oImprovedNoiseClass.singleNoise(xPlane, yPlane, zPlane, aOffset[0], aAmplitude[0], absolute, c1, c2, oWriteableBitmap);

            }
            else
            {
                if (oNoiseBlenderEffect.absolute != absolute)
                {
                    oNoiseBlenderEffect = new NoiseBlenderEffect(octaves - 1, absolute);
                    oNoiseBlenderEffect.aInput = aImageBrush;
                }
                oNoiseBlenderEffect.c1 = c1;
                oNoiseBlenderEffect.c2 = c2;
                oNoiseBlenderEffect.aAmplitude = aAmplitude;
                for (int i = 0; i < octaves; i++)
                {
                    oImprovedNoiseClass.noise(xPlane, yPlane, zPlane, aMult[i], aOffset[i], absolute, aWriteableBitmap[i]);
                }
                oImprovedNoiseClass.currentUIElement.Effect = oNoiseBlenderEffect;
                oWriteableBitmap.Render(oImprovedNoiseClass.currentUIElement, oTransform);
                oWriteableBitmap.Invalidate();

            }
        }

        /// <summary>
        /// calculates 3D improved Perlin noise amplitude*cos(mult*xyz+sum(|noise|)+shift) with a random seed
        /// </summary>
        /// <param name="xPlane">x values of the object with the format r.gba</param>
        /// <param name="yPlane">y values of the object with the format r.gba</param>
        /// <param name="zPlane">z values of the object with the format r.gba</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="aMult">the factors of each noise octave</param>
        /// <param name="aOffset">offset of each noise octave</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="amplitude"></param>
        /// <param name="mult"></param>
        /// <param name="shift"></param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void improvedNoises(ImageBrush xPlane, ImageBrush yPlane, ImageBrush zPlane, int octaves, Vector3d [] aMult,Vector3d[] aOffset, float[] aAmplitude, float amplitude,Vector3d mult, float shift, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            improvedNoises(xPlane, yPlane, zPlane, octaves, (new Random()).Next(), aMult, aOffset, aAmplitude, amplitude, mult, shift, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates 3D improved Perlin noise amplitude*cos(mult*xyz+sum(|noise|)+shift) with a defined seed
        /// </summary>
        /// <param name="xPlane">x values of the object with the format r.gba</param>
        /// <param name="yPlane">y values of the object with the format r.gba</param>
        /// <param name="zPlane">z values of the object with the format r.gba</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="seed"></param>
        /// <param name="aMult">the factors of each noise octave</param>
        /// <param name="aOffset">offset of each noise octave</param>
        /// <param name="aAmplitude">amplitudes of the noise octaves</param>
        /// <param name="amplitude"></param>
        /// <param name="mult"></param>
        /// <param name="shift"></param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void improvedNoises(ImageBrush xPlane, ImageBrush yPlane, ImageBrush zPlane, int octaves, int seed, Vector3d[] aMult, Vector3d[] aOffset, float[] aAmplitude, float amplitude, Vector3d mult, float shift, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset.Length != octaves) return;
            if (aAmplitude.Length != octaves) return;
            if (octaves > 1 && aMult.Length != octaves) return;
            bool isNotEqualSize = oImprovedNoiseClass == null ? true : !(new Size(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight)).Equals(oImprovedNoiseClass.thisSize);
            if (octaves != this.octaves || isNotEqualSize)
            {
                this.octaves = octaves;

                oNoiseBlenderCosineEffect3D = new NoiseBlenderCosineEffect3D(octaves - 1);
                aWriteableBitmap = new WriteableBitmap[octaves];
                aImageBrush = new ImageBrush[octaves];
                for (int i = 0; i < octaves; i++)
                {
                    aWriteableBitmap[i] = new WriteableBitmap(oWriteableBitmap.PixelWidth, oWriteableBitmap.PixelHeight);
                    aImageBrush[i] = new ImageBrush();
                    aImageBrush[i].ImageSource = aWriteableBitmap[i];
                }
                oNoiseBlenderCosineEffect3D.xPlane = xPlane;
                oNoiseBlenderCosineEffect3D.yPlane = yPlane;
                oNoiseBlenderCosineEffect3D.zPlane = zPlane;
           }


            if (oImprovedNoiseClass == null)
            {
                this.seed = seed;
                oImprovedNoiseClass = new ImprovedNoiseClass(seed);
                oImprovedNoiseClass.customUIElement = iUIElement;
                oImprovedNoiseClass.oCustomTransform = oTransform;
            }
            else if (this.seed != seed)
            {
                this.seed = seed;
                oImprovedNoiseClass.reInitialize(seed);

            }

            oNoiseBlenderCosineEffect3D.offsetXY = new Point(aOffset[0].x, aOffset[0].y);
            oNoiseBlenderCosineEffect3D.offsetZ = (float)aOffset[0].z;
            oNoiseBlenderCosineEffect3D.multXY = new Point(aMult[0].x, aMult[0].y);
            oNoiseBlenderCosineEffect3D.multZ = (float)aMult[0].z;
            oNoiseBlenderCosineEffect3D.aInput = aImageBrush;
            oNoiseBlenderCosineEffect3D.c1 = c1;
            oNoiseBlenderCosineEffect3D.c2 = c2;
            oNoiseBlenderCosineEffect3D.aAmplitude = aAmplitude;
            oNoiseBlenderCosineEffect3D.amplitude = 0.5f * amplitude;
            oNoiseBlenderCosineEffect3D.xyMult = new Point(mult.x, mult.y);
            oNoiseBlenderCosineEffect3D.zMult = (float)mult.z;
            oNoiseBlenderCosineEffect3D.shift = shift;
            for (int i = 0; i < octaves; i++)
            {
                oImprovedNoiseClass.noise(xPlane, yPlane, zPlane, aMult[i], aOffset[i], true, aWriteableBitmap[i]);
            }
            oImprovedNoiseClass.currentUIElement.Effect = oNoiseBlenderCosineEffect3D;
            oWriteableBitmap.Render(oImprovedNoiseClass.currentUIElement, oTransform);
            oWriteableBitmap.Invalidate();
        }

        /// <summary>
        /// calculates a 3D improved Perlin noise turbulence sum(noise) or sum(|noise|) for custom objects and a random seed
        /// </summary>
        /// <param name="xPlane">x values of the object with the format r.gba</param>
        /// <param name="yPlane">y values of the object with the format r.gba</param>
        /// <param name="zPlane">z values of the object with the format r.gba</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="aOffset">offset of each noise octave, can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void improvedTurbulence(ImageBrush xPlane, ImageBrush yPlane, ImageBrush zPlane, int octaves, Vector3d[] aOffset, float amplitude, float persistence, bool absolute, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            improvedTurbulence(xPlane, yPlane, zPlane, octaves, (new Random()).Next(),aOffset, amplitude, persistence, absolute, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates a 3D improved Perlin noise turbulence sum(noise) or sum(|noise|) for custom objects and a defined seed
        /// </summary>
        /// <param name="xPlane">x values of the object with the format r.gba</param>
        /// <param name="yPlane">y values of the object with the format r.gba</param>
        /// <param name="zPlane">z values of the object with the format r.gba</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">offset of each noise octave, can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="absolute">noise or |noise|</param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void improvedTurbulence(ImageBrush xPlane, ImageBrush yPlane, ImageBrush zPlane, int octaves, int seed, Vector3d[] aOffset, float amplitude, float persistence, bool absolute, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset == null) aOffset = new Vector3d[octaves];

            float[] aAmplitudes = new float[octaves];
            float value = 1;
            Vector3d[] aMult = new Vector3d[octaves];
            for (int i = 0; i < octaves; i++)
            {
                if (aOffset[i] == null) aOffset[i] = new Vector3d(5000, 5000, 5000);
                aMult[i] = new Vector3d(value, value, value);
                value *= 2f;
                aAmplitudes[i] = amplitude;
                amplitude *= persistence;
            }
            improvedNoises(xPlane, yPlane, zPlane, octaves, seed, aMult, aOffset, aAmplitudes, absolute, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates a 3D improved Perlin noise amplitudeCos*cos(mult*xyz+sum(|noise|)+shift) with a random seed
        /// </summary>
        /// <param name="xPlane">x values of the object with the format r.gba</param>
        /// <param name="yPlane">y values of the object with the format r.gba</param>
        /// <param name="zPlane">z values of the object with the format r.gba</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="aOffset">offset of each noise octave, can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="amplitudeCos"></param>
        /// <param name="mult"></param>
        /// <param name="shift"></param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void improvedTurbulence(ImageBrush xPlane, ImageBrush yPlane, ImageBrush zPlane, int octaves, Vector3d[] aOffset, float amplitude, float persistence, float amplitudeCos, Vector3d mult, float shift, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            improvedTurbulence(xPlane, yPlane, zPlane, octaves,  (new Random()).Next(), aOffset, amplitude, persistence, amplitudeCos, mult, shift, c1, c2, oWriteableBitmap);
        }

        /// <summary>
        /// calculates a 3D improved Perlin noise amplitudeCos*cos(mult*xyz+sum(|noise|)+shift) with a defined seed
        /// </summary>
        /// <param name="xPlane">x values of the object with the format r.gba</param>
        /// <param name="yPlane">y values of the object with the format r.gba</param>
        /// <param name="zPlane">z values of the object with the format r.gba</param>
        /// <param name="octaves">number of single noises</param>
        /// <param name="seed">seed for the permutations</param>
        /// <param name="aOffset">offset of each noise octave, can be null or should match the octaves count</param>
        /// <param name="amplitude">amplitude of the first octave</param>
        /// <param name="persistence"></param>
        /// <param name="amplitudeCos"></param>
        /// <param name="mult"></param>
        /// <param name="shift"></param>
        /// <param name="c1">start color</param>
        /// <param name="c2">end color</param>
        /// <param name="oWriteableBitmap">destination bitmap</param>
        public void improvedTurbulence(ImageBrush xPlane, ImageBrush yPlane, ImageBrush zPlane, int octaves, int seed, Vector3d[] aOffset, float amplitude, float persistence, float amplitudeCos,Vector3d mult ,float shift, Color c1, Color c2, WriteableBitmap oWriteableBitmap)
        {
            if (octaves <= 0 || octaves > 16) return;
            if (aOffset == null) aOffset = new Vector3d[octaves];
            
            float[] aAmplitudes = new float[octaves];
            float value = 1;
            Vector3d[] aMult = new Vector3d[octaves];
            for (int i = 0; i < octaves; i++)
            {
                if (aOffset[i] == null) aOffset[i] = new Vector3d(5000, 5000,5000);
                aMult[i] = new Vector3d(value,value,value);
                value *= 2f;
                aAmplitudes[i] = amplitude;
                amplitude *= persistence;
            }
            improvedNoises(xPlane,yPlane,zPlane,octaves,seed, aMult, aOffset,aAmplitudes, amplitudeCos, mult,shift, c1, c2, oWriteableBitmap);
        }


        #endregion
    }
}
